# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.geometry.polyhedron
r"""
Generic Backend for SDP solvers

This class only lists the methods that should be defined by any
interface with a SDP Solver. All these methods immediately raise
:exc:`NotImplementedError` exceptions when called, and are obviously
meant to be replaced by the solver-specific method. This file can also
be used as a template to create a new interface : one would only need
to replace the occurrences of ``"Nonexistent_SDP_solver"`` by the
solver's name, and replace ``GenericSDPBackend`` by
``SolverName(GenericSDPBackend)`` so that the new solver extends this
class.

AUTHORS:

- Ingolfur Edvardsson (2014-07): initial implementation
"""

#*****************************************************************************
#       Copyright (C) 2014 Ingolfur Edvardsson <ingolfured@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

cdef class GenericSDPBackend:

    cpdef base_ring(self):
        """
        The base ring.

        TESTS::

            sage: from sage.numerical.backends.generic_sdp_backend import GenericSDPBackend
            sage: GenericSDPBackend().base_ring()
            Real Double Field
        """
        from sage.rings.real_double import RDF
        return RDF

    cpdef zero(self):
        """
        Zero of the base ring.

        TESTS::

            sage: from sage.numerical.backends.generic_sdp_backend import GenericSDPBackend
            sage: GenericSDPBackend().zero()
            0.0
        """
        return self.base_ring().zero()

    cpdef int add_variable(self, obj=0.0, name=None) except -1:
        """
        Add a variable.

        This amounts to adding a new column to the matrix. By default,
        the variable is both positive and real.

        INPUT:

        - ``obj`` -- (optional) coefficient of this variable in the objective
          function (default: 0.0)

        - ``name`` -- an optional name for the newly added variable (default:
          ``None``)

        OUTPUT: the index of the newly created variable

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.ncols()
            0
            sage: p.add_variable()
            0
            sage: p.ncols()
            1
            sage: p.add_variable(name='x', obj=1.0)
            3
            sage: p.col_name(3)
            'x'
            sage: p.objective_coefficient(3)
            1.0
        """
        raise NotImplementedError()

    cpdef int add_variables(self, int n, names=None) except -1:
        """
        Add ``n`` variables.

        This amounts to adding new columns to the matrix. By default,
        the variables are both positive and real.

        INPUT:

        - ``n`` -- the number of new variables (must be > 0)

        - ``obj`` -- coefficient of all variables in the objective function (default: 0.0)

        - ``names`` -- list of names (default: ``None``)

        OUTPUT: the index of the variable created last

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.ncols()
            0
            sage: p.add_variables(5)
            4
            sage: p.ncols()
            5
            sage: p.add_variables(2, lower_bound=-2.0, integer=True, names=['a','b'])
            6
        """
        raise NotImplementedError()

    cpdef set_sense(self, int sense):
        """
        Set the direction (maximization/minimization).

        INPUT:

        - ``sense`` -- integer:

          * `+1` => Maximization
          * `-1` => Minimization

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.is_maximization()
            True
            sage: p.set_sense(-1)
            sage: p.is_maximization()
            False
        """
        raise NotImplementedError()

    cpdef  objective_coefficient(self, int variable, coeff=None):
        """
        Set or get the coefficient of a variable in the objective
        function

        INPUT:

        - ``variable`` -- integer; the variable's id

        - ``coeff`` -- double; its coefficient

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variable()
            1
            sage: p.objective_coefficient(0)
            0.0
            sage: p.objective_coefficient(0,2)
            sage: p.objective_coefficient(0)
            2.0
        """
        raise NotImplementedError()

    cpdef  set_objective(self, list coeff, d=0.0):
        """
        Set the objective function.

        INPUT:

        - ``coeff`` -- list of real values, whose i-th element is the
          coefficient of the i-th variable in the objective function

        - ``d`` -- double; the constant term in the linear function (set to `0`
          by default)

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(5)
            5
            sage: p.set_objective([1, 1, 2, 1, 3])
            sage: [p.objective_coefficient(x) for x in range(5)]
            [1.0, 1.0, 2.0, 1.0, 3.0]

        Constants in the objective function are respected.
        """
        raise NotImplementedError()

    cpdef add_linear_constraint(self, coefficients, name=None):
        """
        Add a linear constraint.

        INPUT:

        - ``coefficients`` an iterable with ``(c,v)`` pairs where ``c``
          is a variable index (integer) and ``v`` is a value (real
          value).

        - ``lower_bound`` -- a lower bound, either a real value or ``None``

        - ``upper_bound`` -- an upper bound, either a real value or ``None``

        - ``name`` -- an optional name for this row (default: ``None``)

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(5)
            4
            sage: p.add_linear_constraint(zip(range(5), range(5)), 2.0, 2.0)
            sage: p.row(0)
            ([4, 3, 2, 1], [4.0, 3.0, 2.0, 1.0])                   # optional - Nonexistent_LP_solver
            sage: p.row_bounds(0)
            (2.0, 2.0)
            sage: p.add_linear_constraint( zip(range(5), range(5)), 1.0, 1.0, name='foo')
            sage: p.row_name(-1)
            "foo"
        """
        raise NotImplementedError()

    cpdef add_linear_constraints(self, int number, names=None):
        """
        Add constraints.

        INPUT:

        - ``number`` -- integer; the number of constraints to add

        - ``lower_bound`` -- a lower bound, either a real value or ``None``

        - ``upper_bound`` -- an upper bound, either a real value or ``None``

        - ``names`` -- an optional list of names (default: ``None``)

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(5)
            5
            sage: p.add_linear_constraints(5, None, 2)
            sage: p.row(4)
            ([], [])
            sage: p.row_bounds(4)
            (None, 2.0)
        """
        raise NotImplementedError()

    cpdef int solve(self) except -1:
        """
        Solve the problem.

        .. NOTE::

            This method raises :class:`SDPSolverException` exceptions when
            the solution cannot be computed for any reason (none
            exists, or the LP solver was not able to find it, etc...)

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_linear_constraints(5, 0, None)
            sage: p.add_col(range(5), range(5))
            sage: p.solve()
            0
            sage: p.objective_coefficient(0,1)
            sage: p.solve()
            Traceback (most recent call last):
            ...
            SDPSolverException: ...
        """
        raise NotImplementedError()

    cpdef get_objective_value(self):
        """
        Return the value of the objective function.

        .. NOTE::

           Behaviour is undefined unless ``solve`` has been called before.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(2)
            2
            sage: p.add_linear_constraint([(0,1), (1,2)], None, 3)
            sage: p.set_objective([2, 5])
            sage: p.solve()
            0
            sage: p.get_objective_value()
            7.5
            sage: p.get_variable_value(0)
            0.0
            sage: p.get_variable_value(1)
            1.5
        """
        raise NotImplementedError()

    cpdef get_variable_value(self, int variable):
        """
        Return the value of a variable given by the solver.

        .. NOTE::

           Behaviour is undefined unless ``solve`` has been called before.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(2)
            2
            sage: p.add_linear_constraint([(0,1), (1, 2)], None, 3)
            sage: p.set_objective([2, 5])
            sage: p.solve()
            0
            sage: p.get_objective_value()
            7.5
            sage: p.get_variable_value(0)
            0.0
            sage: p.get_variable_value(1)
            1.5
        """

        raise NotImplementedError()

    cpdef int ncols(self) noexcept:
        """
        Return the number of columns/variables.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.ncols()
            0
            sage: p.add_variables(2)
            2
            sage: p.ncols()
            2
        """

        raise NotImplementedError()

    cpdef int nrows(self) noexcept:
        """
        Return the number of rows/constraints.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.nrows()
            0
            sage: p.add_linear_constraints(2, 2.0, None)
            sage: p.nrows()
            2
        """
        raise NotImplementedError()

    cpdef bint is_maximization(self) noexcept:
        """
        Test whether the problem is a maximization

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.is_maximization()
            True
            sage: p.set_sense(-1)
            sage: p.is_maximization()
            False
        """
        raise NotImplementedError()

    cpdef problem_name(self, name=None):
        """
        Return or define the problem's name.

        INPUT:

        - ``name`` -- string; the problem's name. When set to
          ``NULL`` (default), the method returns the problem's name.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.problem_name("There once was a french fry")
            sage: print(p.problem_name())
            There once was a french fry
        """

        raise NotImplementedError()

    cpdef row(self, int i):
        """
        Return a row.

        INPUT:

        - ``index`` -- integer; the constraint's id

        OUTPUT:

        A pair ``(indices, coeffs)`` where ``indices`` lists the
        entries whose coefficient is nonzero, and to which ``coeffs``
        associates their coefficient on the model of the
        ``add_linear_constraint`` method.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variables(5)
            5
            sage: p.add_linear_constraint(zip(range(5), range(5)), 2, 2)
            sage: p.row(0)
            ([4, 3, 2, 1], [4.0, 3.0, 2.0, 1.0])
            sage: p.row_bounds(0)
            (2.0, 2.0)
        """
        raise NotImplementedError()

    cpdef row_name(self, int index):
        """
        Return the ``index``-th row name.

        INPUT:

        - ``index`` -- integer; the row's id

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_linear_constraints(1, 2, None, name="Empty constraint 1")
            sage: p.row_name(0)
            'Empty constraint 1'
        """
        raise NotImplementedError()

    cpdef col_name(self, int index):
        """
        Return the ``index``-th col name.

        INPUT:

        - ``index`` -- integer; the col's id

        - ``name`` -- (``char *``) its name; when set to ``NULL``
          (default), the method returns the current name

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.add_variable(name="I am a variable")
            1
            sage: p.col_name(0)
            'I am a variable'
        """
        raise NotImplementedError()

    cpdef dual_variable(self, int i, sparse=False):
        """
        The `i`-th dual variable.

        Available after ``self.solve()`` is called, otherwise the result is undefined

        - ``index`` -- integer; the constraint's id

        OUTPUT: the matrix of the `i`-th dual variable

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: p = SemidefiniteProgram(maximization=False, solver="Nonexistent_LP_solver")
            sage: x = p.new_variable()
            sage: p.set_objective(x[0] - x[1])
            sage: a1 = matrix([[1, 2.], [2., 3.]])
            sage: a2 = matrix([[3, 4.], [4., 5.]])
            sage: a3 = matrix([[5, 6.], [6., 7.]])
            sage: b1 = matrix([[1, 1.], [1., 1.]])
            sage: b2 = matrix([[2, 2.], [2., 2.]])
            sage: b3 = matrix([[3, 3.], [3., 3.]])
            sage: p.add_constraint(a1*x[0] + a2*x[1] <= a3)
            sage: p.add_constraint(b1*x[0] + b2*x[1] <= b3)
            sage: p.solve()
            -3.0
            sage: B = p.get_backend()
            sage: x = p.get_values(x).values()
            sage: -(a3*B.dual_variable(0)).trace()-(b3*B.dual_variable(1)).trace()
            -3.0
            sage: g = sum((B.slack(j)*B.dual_variable(j)).trace() for j in range(2)); g
            0.0

        TESTS::

            sage: B.dual_variable(7)  # optional - Nonexistent_LP_solver
            ...
            Traceback (most recent call last):
            ...
            IndexError: list index out of range
            sage: abs(g - B._get_answer()['gap'])  # optional - Nonexistent_LP_solver # tol 1e-22
            0.0
        """
        raise NotImplementedError()

    cpdef slack(self, int i, sparse=False):
        """
        Slack of the `i`-th constraint.

        Available after ``self.solve()`` is called, otherwise the result is
        undefined.

        - ``index`` -- integer; the constraint's id

        OUTPUT: the matrix of the slack of the `i`-th constraint

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: p = SemidefiniteProgram(maximization=False, solver="Nonexistent_LP_solver")
            sage: x = p.new_variable()
            sage: p.set_objective(x[0] - x[1])
            sage: a1 = matrix([[1, 2.], [2., 3.]])
            sage: a2 = matrix([[3, 4.], [4., 5.]])
            sage: a3 = matrix([[5, 6.], [6., 7.]])
            sage: b1 = matrix([[1, 1.], [1., 1.]])
            sage: b2 = matrix([[2, 2.], [2., 2.]])
            sage: b3 = matrix([[3, 3.], [3., 3.]])
            sage: p.add_constraint(a1*x[0] + a2*x[1] <= a3)
            sage: p.add_constraint(b1*x[0] + b2*x[1] <= b3)
            sage: p.solve()
            -3.0
            sage: B = p.get_backend()
            sage: B1 = B.slack(1); B1
            [0.0 0.0]
            [0.0 0.0]
            sage: B1.is_positive_definite()
            True
            sage: x = p.get_values(x).values()
            sage: x[0]*b1 + x[1]*b2 - b3 + B1
            [0.0 0.0]
            [0.0 0.0]

        TESTS::

            sage: B.slack(7)  # optional - Nonexistent_LP_solver
            ...
            Traceback (most recent call last):
            ...
            IndexError: list index out of range
        """
        raise NotImplementedError()

    cpdef solver_parameter(self, name, value=None):
        """
        Return or define a solver parameter.

        INPUT:

        - ``name`` -- string; the parameter

        - ``value`` -- the parameter's value if it is to be defined,
          or ``None`` (default) to obtain its current value

        .. NOTE::

           The list of available parameters is available at
           :meth:`~sage.numerical.sdp.SemidefiniteProgram.solver_parameter`.

        EXAMPLES::

            sage: # optional - nonexistent_lp_solver
            sage: from sage.numerical.backends.generic_sdp_backend import get_solver
            sage: p = get_solver(solver="Nonexistent_LP_solver")
            sage: p.solver_parameter("timelimit")
            sage: p.solver_parameter("timelimit", 60)
            sage: p.solver_parameter("timelimit")
        """
        raise NotImplementedError()


default_solver = None


def default_sdp_solver(solver=None):
    """
    Return/set the default SDP solver used by Sage.

    INPUT:

    - ``solver`` -- one of the following:

      - the string ``'CVXOPT'``, to make the use of the CVXOPT solver
        (see the `CVXOPT <http://cvxopt.org/>`_ web site) the default;

      - a subclass of
        :class:`sage.numerical.backends.generic_sdp_backend.GenericSDPBackend`,
        to make it the default; or

      - ``None`` -- (default) in which case the current default solver
        (a string or a class) is returned

    OUTPUT:

    This function returns the current default solver (a string or a
    class) if ``solver = None`` (default). Otherwise, it sets the
    default solver to the one given. If this solver does not exist, or
    is not available, a :exc:`ValueError` exception is raised.

    EXAMPLES::

        sage: former_solver = default_sdp_solver()
        sage: default_sdp_solver("Cvxopt")
        sage: default_sdp_solver()
        'Cvxopt'
        sage: default_sdp_solver("Yeahhhhhhhhhhh")
        Traceback (most recent call last):
        ...
        ValueError: 'solver' should be set to ...
        sage: default_sdp_solver(former_solver)
        sage: from sage.numerical.backends.generic_sdp_backend import GenericSDPBackend
        sage: class my_sdp_solver(GenericSDPBackend): pass
        sage: default_sdp_solver(my_sdp_solver)
        sage: default_sdp_solver() is my_sdp_solver
        True
    """
    global default_solver

    if solver is None:

        if default_solver is not None:
            return default_solver

        else:
            for s in ["Cvxopt"]:
                try:
                    default_sdp_solver(s)
                    return s
                except ValueError:
                    pass

            from warnings import warn
            warn("default_sdp_solver set to 'Matrix' (MatrixSDPBackend), which can construct but not solve problems. Install cvxopt for actual solver functionality")
            default_sdp_solver("Matrix")

    if callable(solver):
        default_solver = solver
        return

    solver = solver.capitalize()

    if solver == "Cvxopt":
        try:
            from sage.numerical.backends.cvxopt_sdp_backend import CVXOPTSDPBackend
            default_solver = solver
        except ImportError:
            raise ValueError("CVXOPT is not available. Please refer to the documentation to install it.")
    elif solver == "Matrix":
        default_solver = solver

    else:
        raise ValueError("'solver' should be set to 'CVXOPT', 'Matrix', a class, or None.")


cpdef GenericSDPBackend get_solver(solver=None, base_ring=None):
    """
    Return a solver according to the given preferences.

    INPUT:

    - ``solver`` -- one of the following:

      - the string ``'CVXOPT'``, designating the use of the CVXOPT solver
        (see the `CVXOPT <http://cvxopt.org/>`_ web site);

      - a subclass of
        :class:`sage.numerical.backends.generic_sdp_backend.GenericSDPBackend`

      - ``None`` -- (default) in which case the default solver is used (see
        :func:`default_sdp_solver`)

    .. SEEALSO::

        - :func:`default_sdp_solver` -- returns/sets the default SDP solver

    EXAMPLES::

        sage: from sage.numerical.backends.generic_sdp_backend import get_solver
        sage: p = get_solver()

    Passing a class::

        sage: from sage.numerical.backends.generic_sdp_backend import GenericSDPBackend
        sage: class MockSDPBackend(GenericSDPBackend):
        ....:     def solve(self):
        ....:         raise RuntimeError("SDP is too slow")
        sage: P = SemidefiniteProgram(solver=MockSDPBackend)
        sage: P.solve()
        Traceback (most recent call last):
        ...
        RuntimeError: SDP is too slow
    """
    if solver is None:
        solver = default_sdp_solver()

    if callable(solver):
        return solver(base_ring=base_ring)

    solver = solver.capitalize()

    if solver == "Cvxopt":
        from sage.numerical.backends.cvxopt_sdp_backend import CVXOPTSDPBackend
        return CVXOPTSDPBackend(base_ring=base_ring)
    elif solver == "Matrix":
        from sage.numerical.backends.matrix_sdp_backend import MatrixSDPBackend
        return MatrixSDPBackend(base_ring=base_ring)
    else:
        raise ValueError("'solver' should be set to 'CVXOPT', 'Matrix', a class, or None (in which case the default one is used).")



# sage_setup: distribution = sagemath-categories
r"""
Parallelization control

This module defines the singleton class :class:`Parallelism` to govern the
parallelization of computations in some specific topics. It allows the user to
set the number of processes to be used for parallelization.

Some examples of use are provided in the documentation of
:meth:`sage.tensor.modules.comp.Components.contract`.

AUTHORS:

- Marco Mancini, Eric Gourgoulhon, Michal Bejger (2015): initial version
"""

# *****************************************************************************
#       Copyright (C) 2015 Marco Mancini <marco.mancini@obspm.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.structure.sage_object import SageObject
from sage.misc.fast_methods import Singleton
from sage.parallel.ncpus import ncpus
from sage.rings.integer import Integer


class Parallelism(Singleton, SageObject):
    r"""
    Singleton class for managing the number of processes used in parallel
    computations involved in various fields.

    EXAMPLES:

    The number of processes is initialized to 1 (no parallelization) for
    each field::

        sage: Parallelism()
        Number of processes for parallelization:
         - linbox computations: 1
         - tensor computations: 1

    Using 4 processes to parallelize tensor computations::

        sage: Parallelism().set('tensor', nproc=4)
        sage: Parallelism()
        Number of processes for parallelization:
         - linbox computations: 1
         - tensor computations: 4
        sage: Parallelism().get('tensor')
        4

    Using 6 processes to parallelize all types of computations::

        sage: Parallelism().set(nproc=6)
        sage: Parallelism()
        Number of processes for parallelization:
         - linbox computations: 6
         - tensor computations: 6

    Using all the cores available on the computer to parallelize tensor
    computations::

        sage: Parallelism().set('tensor')
        sage: Parallelism()  # random (depends on the computer)
        Number of processes for parallelization:
         - linbox computations: 1
         - tensor computations: 8

    Using all the cores available on the computer to parallelize all types
    of computations::

        sage: Parallelism().set()
        sage: Parallelism()  # random (depends on the computer)
        Number of processes for parallelization:
         - linbox computations: 8
         - tensor computations: 8

    Switching off all parallelizations::

        sage: Parallelism().set(nproc=1)
    """
    def __init__(self):
        r"""
        Construct the single instance of class Parallelism (singleton model).

        TESTS::

            sage: par = Parallelism()
            sage: par
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 1

        Test of the singleton character::

            sage: Parallelism() is par
            True

        The test suite is passed::

            sage: TestSuite(par).run()
        """
        self._default = ncpus()
        # default number of proc. used in parallelizations

        self._nproc = {'tensor': 1, 'linbox': 1}
        # dict. of number of processes to be used
        # (keys: computational field)

    def _repr_(self):
        r"""
        String representation of the object.

        TESTS::

            sage: Parallelism()._repr_()
            'Number of processes for parallelization:\n - linbox computations: 1\n - tensor computations: 1'
        """
        resu = "Number of processes for parallelization:\n"
        for field in sorted(self._nproc):
            resu += " - {} computations: {}\n".format(field, self._nproc[field])
        return resu[:-1]

    def reset(self):
        r"""
        Put the singleton object ``Parallelism()`` in the same state as
        immediately after its creation.

        EXAMPLES:

        State of ``Parallelism()`` just after its creation::

            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 1
            sage: Parallelism().get_default()  # random (depends on the computer)
            8

        Changing some values::

            sage: Parallelism().set_default(6)
            sage: Parallelism().set()
            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 6
             - tensor computations: 6
            sage: Parallelism().get_default()
            6

        Back to the initial state::

            sage: Parallelism().reset()
            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 1
            sage: Parallelism().get_default()  # random (depends on the computer)
            8
        """
        self._default = ncpus()
        for field in self._nproc:
            self._nproc[field] = 1

    def set(self, field=None, nproc=None):
        r"""
        Set the number of processes to be launched for parallel computations
        regarding some specific field.

        INPUT:

        - ``field`` -- (default: ``None``) string specifying the computational
          field for which the number of parallel processes is to be set; if
          ``None``, all fields are considered
        - ``nproc`` -- (default: ``None``) number of processes to be used for
          parallelization; if ``None``, the number of processes will be set to
          the default value, which, unless redefined by :meth:`set_default`,
          is the total number of cores found on the computer.

        EXAMPLES:

        The default is a single processor (no parallelization)::

            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 1

        Asking for parallelization on 4 cores in tensor algebra::

            sage: Parallelism().set('tensor', nproc=4)
            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 4

        Using all the cores available on the computer::

            sage: Parallelism().set('tensor')
            sage: Parallelism()  # random (depends on the computer)
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 8

        Using 6 cores in all parallelizations::

            sage: Parallelism().set(nproc=6)
            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 6
             - tensor computations: 6

        Using all the cores available on the computer in all parallelizations::

            sage: Parallelism().set()
            sage: Parallelism()  # random (depends on the computer)
            Number of processes for parallelization:
             - linbox computations: 8
             - tensor computations: 8

        Switching off the parallelization::

            sage: Parallelism().set(nproc=1)
            sage: Parallelism()
            Number of processes for parallelization:
             - linbox computations: 1
             - tensor computations: 1
        """
        if field is None:
            for fi in self._nproc:
                self.set(field=fi, nproc=nproc)
        else:
            if field not in self._nproc:
                raise KeyError("entry for field {} is not ".format(field) +
                               "implemented in Parallelism")
            if nproc is None:
                self._nproc[field] = self._default
            else:
                if not isinstance(nproc, (int, Integer)):
                    raise TypeError("nproc must be integer")
                self._nproc[field] = nproc

    def get(self, field):
        r"""
        Return the number of processes which will be used in parallel
        computations regarding some specific field.

        INPUT:

        - ``field`` -- string specifying the part of Sage involved in
          parallel computations

        OUTPUT:

        - number of processes used in parallelization of computations
          pertaining to ``field``

        EXAMPLES:

        The default is a single process (no parallelization)::

            sage: Parallelism().reset()
            sage: Parallelism().get('tensor')
            1

        Asking for parallelization on 4 cores::

            sage: Parallelism().set('tensor', nproc=4)
            sage: Parallelism().get('tensor')
            4
        """
        if field not in self._nproc:
            raise KeyError("entry for field {} is not ".format(field) +
                           "implemented in Parallelism()")
        return self._nproc[field]

    def get_all(self):
        r"""
        Return the number of processes which will be used in parallel
        computations in all fields

        OUTPUT:

        - dictionary of the number of processes, with the computational fields
          as keys

        EXAMPLES::

            sage: Parallelism().reset()
            sage: Parallelism().get_all()
            {'linbox': 1, 'tensor': 1}

        Asking for parallelization on 4 cores::

            sage: Parallelism().set(nproc=4)
            sage: Parallelism().get_all()
            {'linbox': 4, 'tensor': 4}
        """
        return self._nproc

    def set_default(self, nproc=None):
        r"""
        Set the default number of processes to be launched in parallel
        computations.

        INPUT:

        - ``nproc`` -- (default: ``None``) default number of processes;
          if ``None``, the number of processes will be set to the total number
          of cores found on the computer.

        EXAMPLES:

        A priori the default number of process for parallelization is the
        total number of cores found on the computer::

            sage: Parallelism().get_default()  # random (depends on the computer)
            8

        Changing it thanks to ``set_default``::

            sage: Parallelism().set_default(nproc=4)
            sage: Parallelism().get_default()
            4

        Setting it back to the total number of cores available on the computer::

            sage: Parallelism().set_default()
            sage: Parallelism().get_default()  # random (depends on the computer)
            8
        """
        if nproc is None:
            self._default = ncpus()
        else:
            if not isinstance(nproc, (int, Integer)):
                raise TypeError("nproc must be integer")
            self._default = nproc

    def get_default(self):
        r"""
        Return the default number of processes to be launched in parallel
        computations.

        EXAMPLES:

        A priori, the default number of process for parallelization is the
        total number of cores found on the computer::

            sage: Parallelism().reset()
            sage: Parallelism().get_default()  # random (depends on the computer)
            8

        It can be changed via :meth:`set_default`::

            sage: Parallelism().set_default(nproc=4)
            sage: Parallelism().get_default()
            4
        """
        return self._default



# sage_setup: distribution = sagemath-categories
r"""
Topological Spaces
"""
#*****************************************************************************
#  Copyright (C) 2015 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.cartesian_product import CartesianProductsCategory
from sage.categories.covariant_functorial_construction import RegressiveCovariantConstructionCategory
from sage.categories.sets_cat import Sets


class TopologicalSpacesCategory(RegressiveCovariantConstructionCategory):

    _functor_category = "Topological"

    def _repr_object_names(self):
        """
        EXAMPLES::

            sage: Groups().Topological()  # indirect doctest
            Category of topological groups
        """
        return "topological {}".format(self.base_category()._repr_object_names())


class TopologicalSpaces(TopologicalSpacesCategory):
    """
    The category of topological spaces.

    EXAMPLES::

        sage: Sets().Topological()
        Category of topological spaces
        sage: Sets().Topological().super_categories()
        [Category of sets]

    The category of topological spaces defines the topological structure,
    which shall be preserved by morphisms::

        sage: Sets().Topological().additional_structure()
        Category of topological spaces

    TESTS::

        sage: TestSuite(Sets().Topological()).run()
    """
    # We must override the general object because the names don't match
    _base_category_class = (Sets,)

    def _repr_object_names(self):
        """
        EXAMPLES::

            sage: Sets().Topological()  # indirect doctest
            Category of topological spaces
        """
        return "topological spaces"

    class CartesianProducts(CartesianProductsCategory):
        def extra_super_categories(self):
            r"""
            Implement the fact that a (finite) Cartesian product of topological spaces is
            a topological space.

            EXAMPLES::

                sage: from sage.categories.topological_spaces import TopologicalSpaces
                sage: C = TopologicalSpaces().CartesianProducts()
                sage: C.extra_super_categories()
                [Category of topological spaces]
                sage: C.super_categories()
                [Category of Cartesian products of sets, Category of topological spaces]
                sage: C.axioms()
                frozenset()
            """
            return [TopologicalSpaces()]

    class SubcategoryMethods:
        @cached_method
        def Connected(self):
            """
            Return the full subcategory of the connected objects of ``self``.

            EXAMPLES::

                sage: Sets().Topological().Connected()
                Category of connected topological spaces

            TESTS::

                sage: TestSuite(Sets().Topological().Connected()).run()
                sage: Sets().Topological().Connected.__module__
                'sage.categories.topological_spaces'
            """
            return self._with_axiom('Connected')

        @cached_method
        def Compact(self):
            """
            Return the subcategory of the compact objects of ``self``.

            EXAMPLES::

                sage: Sets().Topological().Compact()
                Category of compact topological spaces

            TESTS::

                sage: TestSuite(Sets().Topological().Compact()).run()
                sage: Sets().Topological().Compact.__module__
                'sage.categories.topological_spaces'
            """
            return self._with_axiom('Compact')

    class Connected(CategoryWithAxiom):
        """
        The category of connected topological spaces.
        """

        class CartesianProducts(CartesianProductsCategory):
            def extra_super_categories(self):
                r"""
                Implement the fact that a (finite) Cartesian product of connected
                topological spaces is connected.

                EXAMPLES::

                    sage: from sage.categories.topological_spaces import TopologicalSpaces
                    sage: C = TopologicalSpaces().Connected().CartesianProducts()
                    sage: C.extra_super_categories()
                    [Category of connected topological spaces]
                    sage: C.super_categories()
                    [Category of Cartesian products of topological spaces,
                     Category of connected topological spaces]
                    sage: C.axioms()
                    frozenset({'Connected'})
                """
                return [TopologicalSpaces().Connected()]

    class Compact(CategoryWithAxiom):
        """
        The category of compact topological spaces.
        """

        class CartesianProducts(CartesianProductsCategory):
            def extra_super_categories(self):
                r"""
                Implement the fact that a (finite) Cartesian product of compact
                topological spaces is compact.

                EXAMPLES::

                    sage: from sage.categories.topological_spaces import TopologicalSpaces
                    sage: C = TopologicalSpaces().Compact().CartesianProducts()
                    sage: C.extra_super_categories()
                    [Category of compact topological spaces]
                    sage: C.super_categories()
                    [Category of Cartesian products of topological spaces,
                     Category of compact topological spaces]
                    sage: C.axioms()
                    frozenset({'Compact'})
                """
                return [TopologicalSpaces().Compact()]



# sage_setup: distribution = sagemath-categories
"""
Image Sets
"""

# ****************************************************************************
#       Copyright (C) 2008      Mike Hansen <mhansen@gmail.com>
#                     2012      Christian Stump
#                     2020-2021 Frédéric Chapoton
#                     2021      Travis Scrimshaw
#                     2021-2024 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from collections.abc import Iterator

from sage.categories.map import Map
from sage.categories.poor_man_map import PoorManMap
from sage.categories.sets_cat import Sets
from sage.categories.enumerated_sets import EnumeratedSets
from sage.misc.cachefunc import cached_method
from sage.rings.infinity import Infinity
from sage.rings.integer import Integer
from sage.structure.element import Expression
from sage.structure.parent import Parent

from .set import Set_base, Set_add_sub_operators, Set_boolean_operators


class ImageSubobject(Parent):
    r"""
    The subset defined as the image of another set under a fixed map.

    Let `f: X \to Y` be a function. Then the image of `f` is defined as

    .. MATH::

        \{ f(x) | x \in X \} \subseteq Y.

    INPUT:

    - ``map`` -- a function

    - ``domain_subset`` -- the set `X`; optional if `f` has a domain

    - ``is_injective`` -- whether the ``map`` is injective:
      - ``None`` (default): infer from ``map`` or default to ``False``
      - ``False``: do not assume that ``map`` is injective
      - ``True``: ``map`` is known to be injective
      - ``'check'``: raise an error when ``map`` is not injective

    - ``inverse`` -- a function (optional); a map from `f(X)` to `X`

    EXAMPLES::

        sage: import itertools
        sage: from sage.sets.image_set import ImageSubobject
        sage: D = ZZ
        sage: I = ImageSubobject(abs, ZZ, is_injective='check')
        sage: list(itertools.islice(I, 10))
        Traceback (most recent call last):
        ...
        ValueError: The map <built-in function abs> from Integer Ring is not injective: 1
    """
    def __init__(self, map, domain_subset, *, category=None, is_injective=None, inverse=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: # needs sage.modules
            sage: M = CombinatorialFreeModule(ZZ, [0,1,2,3])
            sage: R.<x,y> = QQ[]
            sage: H = Hom(M, R, category=Sets())
            sage: f = H(lambda v: v[0]*x + v[1]*(x^2-y) + v[2]^2*(y+2) + v[3] - v[0]^2)
            sage: Im = f.image()
            sage: TestSuite(Im).run(skip=['_test_an_element', '_test_pickling',
            ....:                         '_test_some_elements', '_test_elements'])
        """
        if not isinstance(domain_subset, Parent):
            from sage.sets.set import Set
            domain_subset = Set(domain_subset)

        if not isinstance(map, Map) and not isinstance(map, PoorManMap):
            map_name = f"The map {map}"
            if isinstance(map, Expression) and map.is_callable():
                domain = map.parent().base()
                if len(map.arguments()) != 1:
                    from sage.modules.free_module import FreeModule
                    domain = FreeModule(domain, len(map.arguments()))
                function = map

                def map(arg):
                    return function(*arg)
            else:
                domain = domain_subset
            map = PoorManMap(map, domain, name=map_name)

        if isinstance(map, Map):
            map_category = map.category_for()
            if is_injective is None:
                try:
                    is_injective = map.is_injective()
                except NotImplementedError:
                    is_injective = False
        else:
            map_category = Sets()
            if is_injective is None:
                is_injective = False

        if category is None:
            category = map_category._meet_(domain_subset.category())

        category = category.Subobjects()
        if domain_subset in Sets().Finite() or map.codomain() in Sets().Finite():
            category = category.Finite()
        elif is_injective and domain_subset in Sets.Infinite():
            category = category.Infinite()

        if domain_subset in EnumeratedSets():
            category = category & EnumeratedSets()

        Parent.__init__(self, category=category)

        self._map = map
        self._inverse = inverse
        self._domain_subset = domain_subset
        self._is_injective = is_injective

    def __eq__(self, other):
        r"""
        EXAMPLES::

            sage: from sage.sets.image_set import ImageSubobject
            sage: D = ZZ
            sage: def f(x):
            ....:     return 2 * x
            sage: I = ImageSubobject(f, ZZ)
            sage: I == ImageSubobject(f, ZZ)
            True

        This method does not take into account whether an inverse is provided,
        injectivity is declared, or the category::

            sage: def f_inv(y):
            ....:     return y // 2
            sage: I == ImageSubobject(f, ZZ, inverse=f_inv)
            True
            sage: I == ImageSubobject(f, ZZ, is_injective=True)
            True
            sage: I.category()
            Category of enumerated subobjects of sets
            sage: I == ImageSubobject(f, ZZ, category=EnumeratedSets().Infinite())
            True
        """
        if not isinstance(other, ImageSubobject):
            return False
        return (self._map == other._map
                and self._domain_subset == other._domain_subset)

    def __ne__(self, other):
        r"""
        EXAMPLES::

            sage: from sage.sets.image_set import ImageSubobject
            sage: D = ZZ
            sage: def f(x):
            ....:     return 2 * x
            sage: I = ImageSubobject(f, ZZ)
            sage: I != ImageSubobject(f, QQ)
            True
        """
        return not (self == other)

    def __hash__(self):
        r"""
        TESTS::

            sage: from sage.sets.image_set import ImageSubobject
            sage: def f(x):
            ....:     return 2 * x
            sage: I = ImageSubobject(f, ZZ)
            sage: hash(I) == hash(ImageSubobject(f, ZZ))
            True
        """
        return hash((self._map, self._domain_subset))

    def _element_constructor_(self, x):
        """
        EXAMPLES::

            sage: import itertools
            sage: from sage.sets.image_set import ImageSubobject
            sage: D = ZZ
            sage: I = ImageSubobject(lambda x: 2 * x, ZZ, inverse=lambda x: x/2)
            sage: I(8/2)
            4
            sage: _.parent()
            Integer Ring
            sage: I(10/2)
            Traceback (most recent call last):
            ...
            ValueError: 5 is not in Image of Integer Ring by The map <function <lambda> at ...> from Integer Ring
            sage: 6 in I
            True
            sage: 7 in I
            False
        """
        # Same as ImageManifoldSubset.__contains__
        codomain = self._map.codomain()
        if codomain is not None and x not in codomain:
            raise ValueError(f"{x} is not in {self}")
        if self._inverse is not None:
            preimage = self._inverse(x)
            if preimage not in self._domain_subset:
                raise ValueError(f"{x} is not in {self}")
            preimage = self._map.domain()(preimage)
            y = self._map(preimage)
            if y == x:
                return y
            raise ValueError(f"{x} is not in {self}")
        raise NotImplementedError

    def ambient(self):
        """
        Return the ambient set of ``self``, which is the codomain of
        the defining map.

        EXAMPLES::

            sage: # needs sage.modules
            sage: M = CombinatorialFreeModule(QQ, [0, 1, 2, 3])
            sage: R.<x,y> = ZZ[]
            sage: H = Hom(M, R, category=Sets())
            sage: f = H(lambda v: floor(v[0])*x + ceil(v[3] - v[0]^2))
            sage: Im = f.image()
            sage: Im.ambient() is R
            True

            sage: P = Partitions(3).map(attrcall('conjugate'))
            sage: P.ambient() is None
            True

            sage: R = Permutations(10).map(attrcall('reduced_word'))
            sage: R.ambient() is None
            True
        """
        return self._map.codomain()

    def lift(self, x):
        r"""
        Return the lift ``x`` to the ambient space, which is ``x``.

        EXAMPLES::

            sage: # needs sage.modules
            sage: M = CombinatorialFreeModule(QQ, [0, 1, 2, 3])
            sage: R.<x,y> = ZZ[]
            sage: H = Hom(M, R, category=Sets())
            sage: f = H(lambda v: floor(v[0])*x + ceil(v[3] - v[0]^2))
            sage: Im = f.image()
            sage: p = Im.lift(Im.an_element()); p
            2*x - 4
            sage: p.parent() is R
            True
        """
        return x

    def retract(self, x):
        """
        Return the retract of ``x`` from the ambient space, which is ``x``.

        .. WARNING::

            This does not check that ``x`` is actually in the image.

        EXAMPLES::

            sage: # needs sage.modules
            sage: M = CombinatorialFreeModule(QQ, [0, 1, 2, 3])
            sage: R.<x,y> = ZZ[]
            sage: H = Hom(M, R, category=Sets())
            sage: f = H(lambda v: floor(v[0])*x + ceil(v[3] - v[0]^2))
            sage: Im = f.image()
            sage: p = 2 * x - 4
            sage: Im.retract(p).parent()
            Multivariate Polynomial Ring in x, y over Integer Ring
        """
        return x

    def _repr_(self) -> str:
        r"""
        TESTS::

            sage: Partitions(3).map(attrcall('conjugate'))
            Image of Partitions of the integer 3 by
             The map *.conjugate() from Partitions of the integer 3
        """
        return f"Image of {self._domain_subset} by {self._map}"

    @cached_method
    def cardinality(self) -> Integer:
        r"""
        Return the cardinality of ``self``.

        EXAMPLES:

        Injective case (note that
        :meth:`~sage.categories.enumerated_sets.EnumeratedSets.ParentMethods.map`
        defaults to ``is_injective=True``):

            sage: R = Permutations(10).map(attrcall('reduced_word'))
            sage: R.cardinality()
            3628800

            sage: Evens = ZZ.map(lambda x: 2 * x)
            sage: Evens.cardinality()
            +Infinity

        Non-injective case::

            sage: Z7 = Set(range(7))
            sage: from sage.sets.image_set import ImageSet
            sage: Z4711 = ImageSet(lambda x: x**4 % 11, Z7, is_injective=False)
            sage: Z4711.cardinality()
            6

            sage: Squares = ImageSet(lambda x: x^2, ZZ, is_injective=False,
            ....:                    category=Sets().Infinite())
            sage: Squares.cardinality()
            +Infinity

            sage: Mod2 = ZZ.map(lambda x: x % 2, is_injective=False)
            sage: Mod2.cardinality()
            Traceback (most recent call last):
            ...
            NotImplementedError: cannot determine cardinality of a
            non-injective image of an infinite set
        """
        domain_cardinality = self._domain_subset.cardinality()
        if self._is_injective and self._is_injective != 'check':
            return domain_cardinality
        if self in Sets().Infinite():
            return Infinity
        if domain_cardinality == Infinity:
            raise NotImplementedError('cannot determine cardinality of a non-injective image of an infinite set')
        # Fallback like EnumeratedSets.ParentMethods.__len__
        return Integer(len(list(iter(self))))

    def __iter__(self) -> Iterator:
        r"""
        Return an iterator over the elements of ``self``.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: P = Partitions()
            sage: H = Hom(P, ZZ)
            sage: f = H(ZZ.sum)
            sage: X = f.image()
            sage: it = iter(X)
            sage: [next(it) for _ in range(5)]
            [0, 1, 2, 3, 4]
        """
        if self._is_injective and self._is_injective != 'check':
            for x in self._domain_subset:
                yield self._map(x)
        else:
            visited = set()
            for x in self._domain_subset:
                y = self._map(x)
                if y in visited:
                    if self._is_injective == 'check':
                        raise ValueError(f'{self._map} is not injective: {y}')
                    continue
                visited.add(y)
                yield y

    def _an_element_(self):
        r"""
        Return an element of this set.

        EXAMPLES::

            sage: R = SymmetricGroup(10).map(attrcall('reduced_word'))                  # needs sage.groups
            sage: R.an_element()                                                        # needs sage.groups
            [9, 8, 7, 6, 5, 4, 3, 2]
        """
        domain_element = self._domain_subset.an_element()
        return self._map(domain_element)

    def _sympy_(self):
        r"""
        Return an instance of a subclass of SymPy ``Set`` corresponding to ``self``.

        EXAMPLES::

            sage: from sage.sets.image_set import ImageSet
            sage: S = ImageSet(sin, RealSet.open(0, pi/4)); S                           # needs sage.symbolic
            Image of (0, 1/4*pi) by The map sin from (0, 1/4*pi)
            sage: S._sympy_()                                                           # needs sage.symbolic
            ImageSet(Lambda(x, sin(x)), Interval.open(0, pi/4))
        """
        from sympy import imageset
        try:
            sympy_map = self._map._sympy_()
        except AttributeError:
            sympy_map = self._map
        return imageset(sympy_map,
                        self._domain_subset._sympy_())


class ImageSet(ImageSubobject, Set_base, Set_add_sub_operators, Set_boolean_operators):
    r"""
    Image of a set by a map.

    EXAMPLES::

        sage: from sage.sets.image_set import ImageSet

    Symbolics::

        sage: ImageSet(sin, RealSet.open(0, pi/4))                                      # needs sage.symbolic
        Image of (0, 1/4*pi) by The map sin from (0, 1/4*pi)
        sage: _.an_element()                                                            # needs sage.symbolic
        1/2*sqrt(-sqrt(2) + 2)

        sage: # needs sage.symbolic
        sage: sos(x,y) = x^2 + y^2; sos
        (x, y) |--> x^2 + y^2
        sage: ImageSet(sos, ZZ^2)
        Image of
         Ambient free module of rank 2 over the principal ideal domain Integer Ring by
         The map (x, y) |--> x^2 + y^2 from Vector space of dimension 2 over Symbolic Ring
        sage: _.an_element()
        1
        sage: ImageSet(sos, Set([(3, 4), (3, -4)]))
        Image of {...(3, -4)...} by
         The map (x, y) |--> x^2 + y^2 from Vector space of dimension 2 over Symbolic Ring
        sage: _.an_element()
        25
    """
    pass



# sage_setup: distribution = sagemath-categories
r"""
Fast Numerical Evaluation

For many applications such as numerical integration, differential
equation approximation, plotting a 3d surface, optimization problems,
monte-carlo simulations, etc., one wishes to pass around and evaluate
a single algebraic expression many, many times at various floating
point values. Doing this via recursive calls over a python
representation of the object (even if Maxima or other outside packages
are not involved) is extremely inefficient.

The solution implemented in this module, by Robert Bradshaw (2008-10),
has been superseded by :func:`~sage.ext.fast_callable.fast_callable`.
All that remains here is a compatible interface function :func:`fast_float`.

AUTHORS:

- Robert Bradshaw (2008-10): Initial version
"""

#*****************************************************************************
#       Copyright (C) 2008 Robert Bradshaw <robertwb@math.washington.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.ext.fast_callable import fast_callable, Wrapper


def fast_float(f, *vars, expect_one_var=False):
    """
    Try to create a function that evaluates f quickly using
    floating-point numbers, if possible.

    On failure, returns the input unchanged.

    This is an alternative interface to :func:`sage.ext.fast_callable.fast_callable`.
    :issue:`32268` proposes to deprecate this function.

    INPUT:

    - ``f`` -- an expression
    - ``vars`` -- the names of the arguments
    - ``expect_one_var`` -- don't give deprecation warning if ``vars`` is
      omitted, as long as expression has only one var

    EXAMPLES::

        sage: from sage.ext.fast_eval import fast_float
        sage: x,y = var('x,y')                                                          # needs sage.symbolic
        sage: f = fast_float(sqrt(x^2+y^2), 'x', 'y')                                   # needs sage.symbolic
        sage: f(3,4)                                                                    # needs sage.symbolic
        5.0

    Specifying the argument names is essential, as fast_float objects
    only distinguish between arguments by order. ::

        sage: # needs sage.symbolic
        sage: f = fast_float(x-y, 'x','y')
        sage: f(1,2)
        -1.0
        sage: f = fast_float(x-y, 'y','x')
        sage: f(1,2)
        1.0
    """
    if isinstance(f, (tuple, list)):
        return tuple([fast_float(x, *vars, expect_one_var=expect_one_var) for x in f])

    cdef int i
    for i from 0 <= i < len(vars):
        if not isinstance(vars[i], str):
            v = str(vars[i])
            # inexact generators display as 1.00..0*x
            if '*' in v:
                v = v[v.index('*')+1:]
            vars = vars[:i] + (v,) + vars[i+1:]

    try:
        return fast_callable(f, vars=vars, domain=float,
                             expect_one_var=expect_one_var)
    except AttributeError:
        pass

    try:
        from sage.symbolic.ring import SR
        return fast_float(SR(f), *vars)
    except TypeError:
        pass

    if f is None:
        raise TypeError("no way to make fast_float from None")

    return f


def is_fast_float(x):
    return isinstance(x, Wrapper)



# sage_setup: distribution = sagemath-categories
"""
Exact Cover Problem via Dancing Links
"""
# dlx.py
# Copyright (c) 2006,2008 Antti Ajanki <antti.ajanki@iki.fi>

# Heavily Modified Feb 2008, Tom Boothby
#  * Added a function which takes a Sage matrix and solves
#    the exact cover problem for it.
#  * Recursive search is now iterative
#  * Removed callback functionality
#  * Revamped the class to be a pythonic generator; new usage:
#        for cover in DLXMatrix(ones, initialsolution):
#            blah(cover)
#  * DOCUMENTATION AND TESTS GALORE HOLYCRAP 100% COVERAGE!

# DLXMatrix class is used to store and solve an exact cover problem
# with help of Dancing Links [1] technique by Donald Knuth. The
# problem can be stated as an attempt to leave out rows of a 0/1
# matrix until remaining matrix has exactly one 1 in each column.
# Knuth proposes a fast solution technique based on clever trick with
# double linked list.
#
# This implementation will return row numbers of the solution instead
# column indexes (like in the Knuth's paper).
#
# [1] Donald E Knuth, Dancing links, preprint, available at
# http://www-cs-faculty.stanford.edu/~knuth/preprints.html

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

ROOTNODE = 0

# Node's attributes
# LEFT, RIGHT, UP and DOWN are identifiers of corresponding neighbors
# INDEX is the (row) index of the node (None in the header nodes)
# COLUMN/COUNT is column index on a regular node and count of nodes on
# a header node
LEFT = 0
RIGHT = 1
UP = 2
DOWN = 3
COLUMN = 4
INDEX = 5
COUNT = 5


class DLXMatrix:
    def __init__(self, ones, initialsolution=None):
        """
        Solve the Exact Cover problem by using the Dancing Links algorithm
        described by Knuth.

        Consider a matrix M with entries of 0 and 1, and compute a subset
        of the rows of this matrix which sum to the vector of all 1s.

        The dancing links algorithm works particularly well for sparse
        matrices, so the input is a list of lists of the form: (note the
        1-index!)::

          [
           [1, [i_11,i_12,...,i_1r]]
           ...
           [m,[i_m1,i_m2,...,i_ms]]
          ]

        where M[j][i_jk] = 1.

        The first example below corresponds to the matrix::

           1110
           1010
           0100
           0001

        which is exactly covered by::

           1110
           0001

        and

        ::

           1010
           0100
           0001

        EXAMPLES::

            sage: from sage.combinat.dlx import *
            sage: ones = [[1,[1,2,3]]]
            sage: ones+= [[2,[1,3]]]
            sage: ones+= [[3,[2]]]
            sage: ones+= [[4,[4]]]
            sage: DLXM = DLXMatrix(ones,[4])
            sage: for C in DLXM:
            ....:      print(C)
            [4, 1]
            [4, 2, 3]

        .. NOTE::

            The 0 entry is reserved internally for headers in the
            sparse representation, so rows and columns begin their
            indexing with 1.  Apologies for any heartache this
            causes. Blame the original author, or fix it yourself.
        """
        if initialsolution is None:
            initialsolution = []
        self._cursolution = []
        self._nodes = [[0, 0, None, None, None, None]]
        self._constructmatrix(ones, initialsolution)
        self._level = 0
        self._stack = [(None, None)]

    def __eq__(self, other):
        r"""
        Return ``True`` if every attribute of ``other`` matches the attribute
        of ``self``.

        INPUT:

        - ``other`` -- a DLX matrix

        EXAMPLES::

            sage: from sage.combinat.dlx import *
            sage: M = DLXMatrix([[1,[1]]])
            sage: M == loads(dumps(M))
            True
        """
        if not isinstance(other, DLXMatrix):
            return False
        return self.__dict__ == other.__dict__

    def __iter__(self):
        """
        Return ``self``.

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: M = DLXMatrix([[1,[1]]])
            sage: M.__iter__() is M
            True
        """

        return self

    def _walknodes(self, firstnode, direction):
        """
        Generator for iterating over all nodes in given ``direction`` (not
        including ``firstnode``).

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: ones = [[1,[1,2,3]]]
            sage: ones+= [[2,[1,3]]]
            sage: ones+= [[3,[2]]]
            sage: ones+= [[4,[4]]]
            sage: DLX = DLXMatrix(ones,[4])
            sage: count = 0
            sage: for c in DLX._walknodes(ROOTNODE,RIGHT):
            ....:     count += DLX._nodes[c][COUNT]
            ....:     for d in DLX._walknodes(c,DOWN):
            ....:         count -= 1
            sage: count
            0
        """
        nodetable = self._nodes
        n = nodetable[firstnode][direction]
        while n != firstnode:
            yield n
            n = nodetable[n][direction]

    def _constructmatrix(self, ones, initialsolution=None):
        """
        Construct the (sparse) DLX matrix based on list ``'ones'``.

        The first component in the list elements is row index (which
        will be returned by solve) and the second component is list of
        column indexes of ones in given row.

        'initialsolution' is list of row indexes that are required to be
        part of the solution. They will be removed from the matrix.

        .. NOTE::

            Rows and cols are 1-indexed ; the zero index is reserved for
            the root node and column heads.

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: ones = [[1,[1,2,3]]]
            sage: ones+= [[2,[1,3]]]
            sage: ones+= [[3,[2]]]
            sage: ones+= [[4,[4]]]
            sage: DLX = DLXMatrix([[1,[1]]])
            sage: DLX._constructmatrix(ones,[4])
            sage: c = DLX._nodes[ROOTNODE][RIGHT]
            sage: fullcount = 0
            sage: while c != ROOTNODE:
            ....:     fullcount += DLX._nodes[c][COUNT]
            ....:     d = DLX._nodes[c][DOWN]
            ....:     while d != c:
            ....:         bad = DLX._nodes[DLX._nodes[d][DOWN]][UP] != d
            ....:         bad|= DLX._nodes[DLX._nodes[d][UP]][DOWN] != d
            ....:         bad|= DLX._nodes[DLX._nodes[d][LEFT]][RIGHT] != d
            ....:         bad|= DLX._nodes[DLX._nodes[d][RIGHT]][LEFT] != d
            ....:         if bad:
            ....:             raise RuntimeError("linked list inconsistent")
            ....:         d = DLX._nodes[d][DOWN]
            ....:     c = DLX._nodes[c][RIGHT]
            sage: fullcount
            6
        """
        if initialsolution is None:
            initialsolution = []
        self._cursolution = []
        # LEFT, RIGHT, UP, DOWN, COLUMN, INDEX/COUNT
        self._nodes = [[ROOTNODE, ROOTNODE, None, None, None, None]]

        # optimization: local variables are faster
        nodetable = self._nodes
        ones.sort()
        pruneNodes = []
        headers = [ROOTNODE]  # indexes of header nodes for faster access
        for r in ones:
            curRow = r[0]  # row index
            columns = r[1]  # column indexes
            if not (columns):
                continue
            columns.sort()

            # Do we need more headers?
            while len(headers) <= columns[-1]:
                lastheader = headers[-1]
                newind = len(nodetable)
                nodetable.append([lastheader, ROOTNODE, newind, newind, None, 0])
                nodetable[ROOTNODE][LEFT] = newind
                nodetable[lastheader][RIGHT] = newind
                headers.append(newind)

            # Add new nodes to indexes newind..newind+len(columns)-1
            # LEFT and RIGHT links can be calculated when created the
            # node, only UP, DOWN and COUNT have to be updated when
            # adding new nodes
            newind = len(nodetable)
            for i, c in enumerate(columns):
                h = headers[c]
                l = newind + ((i - 1) % len(columns))
                r = newind + ((i + 1) % len(columns))
                nodetable.append([l, r, nodetable[h][UP], h, h, curRow])
                nodetable[nodetable[h][UP]][DOWN] = newind + i
                nodetable[h][UP] = newind + i
                nodetable[h][COUNT] += 1

            if curRow in initialsolution:
                pruneNodes.append(newind)

        # Remove columns that are required to be in the solution
        for n in pruneNodes:
            self._cursolution += [nodetable[n][INDEX]]
            self._covercolumn(nodetable[n][COLUMN])
            for j in self._walknodes(n, RIGHT):
                self._covercolumn(nodetable[j][COLUMN])

    def _covercolumn(self, c):
        """
        Perform the column covering operation, as described by Knuth's
        pseudocode::

           cover(c):
                i <- D[c]
                while i != c:
                    j <- R[i]
                    while j != i
                        D[U[j]] <- D[j]
                        U[D[j]] <- U[j]
                        N[C[j]] <- N[C[j]] - 1
                        j <- R[j]
                    i <- D[i]

        This is undone by the uncover operation.

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: M = DLXMatrix([[1,[1,3]],[2,[1,2]],[3,[2]]])
            sage: one = M._nodes[ROOTNODE][RIGHT]
            sage: M._covercolumn(one)
            sage: two = M._nodes[ROOTNODE][RIGHT]
            sage: three = M._nodes[two][RIGHT]
            sage: M._nodes[three][RIGHT] == ROOTNODE
            True
            sage: M._nodes[two][COUNT]
            1
            sage: M._nodes[three][COUNT]
            0
        """
        nodetable = self._nodes
        nodetable[nodetable[c][RIGHT]][LEFT] = nodetable[c][LEFT]
        nodetable[nodetable[c][LEFT]][RIGHT] = nodetable[c][RIGHT]
        for i in self._walknodes(c, DOWN):
            for j in self._walknodes(i, RIGHT):
                nodetable[nodetable[j][DOWN]][UP] = nodetable[j][UP]
                nodetable[nodetable[j][UP]][DOWN] = nodetable[j][DOWN]
                nodetable[nodetable[j][COLUMN]][COUNT] -= 1

    def _uncovercolumn(self, c):
        """
        Perform the column uncovering operation, as described by Knuth's
        pseudocode::

            uncover(c):
                i <- U[c]
                while i != c:
                    j <- L[i]
                    while j != i
                        U[j] <- U[D[j]]
                        D[j] <- D[U[j]]
                        N[C[j]] <- N[C[j]] + 1
                        j <- L[j]
                    i <- U[i]

        This undoes by the cover operation since everything is done in the
        reverse order.

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: M = DLXMatrix([[1,[1,3]],[2,[1,2]],[3,[2]]])
            sage: one = M._nodes[ROOTNODE][RIGHT]
            sage: M._covercolumn(one)
            sage: two = M._nodes[ROOTNODE][RIGHT]
            sage: M._uncovercolumn(one)
            sage: M._nodes[two][LEFT] == one
            True
            sage: M._nodes[two][COUNT]
            2
        """
        nodetable = self._nodes
        for i in self._walknodes(c, UP):
            for j in self._walknodes(i, LEFT):
                nodetable[nodetable[j][DOWN]][UP] = j
                nodetable[nodetable[j][UP]][DOWN] = j
                nodetable[nodetable[j][COLUMN]][COUNT] += 1
        nodetable[nodetable[c][RIGHT]][LEFT] = c
        nodetable[nodetable[c][LEFT]][RIGHT] = c

    def __next__(self):
        """
        Search for the first solution we can find, and return it.

        Knuth describes the Dancing Links algorithm recursively, though
        actually implementing it as a recursive algorithm is permissible
        only for highly restricted problems. (for example, the original
        author implemented this for Sudoku, and it works beautifully
        there)

        What follows is an iterative description of DLX::

            stack <- [(NULL)]
            level <- 0
            while level >= 0:
                cur <- stack[level]
                if cur = NULL:
                    if R[h] = h:
                        level <- level - 1
                        yield solution
                    else:
                        cover(best_column)
                        stack[level] = best_column
                else if D[cur] != C[cur]:
                    if cur != C[cur]:
                        delete solution[level]
                        for j in L[cur], L[L[cur]], ... , while j != cur:
                            uncover(C[j])
                    cur <- D[cur]
                    solution[level] <- cur
                    stack[level] <- cur
                    for j in R[cur], R[R[cur]], ... , while j != cur:
                        cover(C[j])
                    level <- level + 1
                    stack[level] <- (NULL)
                else:
                    if C[cur] != cur:
                        delete solution[level]
                        for j in L[cur], L[L[cur]], ... , while j != cur:
                            uncover(C[j])
                    uncover(cur)
                    level <- level - 1

        TESTS::

            sage: from sage.combinat.dlx import *
            sage: M = DLXMatrix([[1,[1,2]],[2,[2,3]],[3,[1,3]]])
            sage: while 1:
            ....:     try:
            ....:         C = next(M)
            ....:     except StopIteration:
            ....:         print("StopIteration")
            ....:         break
            ....:     print(C)
            StopIteration
            sage: M = DLXMatrix([[1,[1,2]],[2,[2,3]],[3,[3]]])
            sage: for C in M:
            ....:       print(C)
            [1, 3]
            sage: M = DLXMatrix([[1,[1]],[2,[2,3]],[3,[2]],[4,[3]]])
            sage: for C in M:
            ....:       print(C)
            [1, 2]
            [1, 3, 4]
        """
        nodetable = self._nodes  # optimization: local variables are faster

        while self._level >= 0:
            c, r = self._stack[self._level]
            if c is None:
                if nodetable[ROOTNODE][RIGHT] == ROOTNODE:
                    self._level -= 1
                    return self._cursolution
                else:
                    c = nodetable[ROOTNODE][RIGHT]
                    maxcount = nodetable[nodetable[ROOTNODE][RIGHT]][COUNT]
                    for j in self._walknodes(ROOTNODE, RIGHT):
                        if nodetable[j][COUNT] < maxcount:
                            c = j
                            maxcount = nodetable[j][COUNT]
                    self._covercolumn(c)
                    self._stack[self._level] = (c, c)
            elif nodetable[r][DOWN] != c:
                if c != r:
                    self._cursolution = self._cursolution[:-1]
                    for j in self._walknodes(r, LEFT):
                        self._uncovercolumn(nodetable[j][COLUMN])
                r = nodetable[r][DOWN]
                self._cursolution += [nodetable[r][INDEX]]
                for j in self._walknodes(r, RIGHT):
                    self._covercolumn(nodetable[j][COLUMN])
                self._stack[self._level] = (c, r)
                self._level += 1
                if len(self._stack) == self._level:
                    self._stack.append((None, None))
                else:
                    self._stack[self._level] = (None, None)
            else:
                if c != r:
                    self._cursolution = self._cursolution[:-1]
                    for j in self._walknodes(r, LEFT):
                        self._uncovercolumn(nodetable[j][COLUMN])
                self._uncovercolumn(c)
                self._level -= 1

        raise StopIteration

    next = __next__


def AllExactCovers(M):
    """
    Use A. Ajanki's DLXMatrix class to solve the exact cover
    problem on the matrix M (treated as a dense binary matrix).

    EXAMPLES::

        sage: # needs sage.modules
        sage: M = Matrix([[1,1,0],[1,0,1],[0,1,1]])  # no exact covers
        sage: for cover in AllExactCovers(M):
        ....:     print(cover)
        sage: M = Matrix([[1,1,0],[1,0,1],[0,0,1],[0,1,0]]) # two exact covers
        sage: for cover in AllExactCovers(M):
        ....:     print(cover)
        [(1, 1, 0), (0, 0, 1)]
        [(1, 0, 1), (0, 1, 0)]
    """
    ones = []
    r = 1    # damn 1-indexing
    for R in M.rows():
        row = [i for i, Ri in enumerate(R, start=1) if Ri]
        ones.append([r, row])
        r += 1
    for s in DLXMatrix(ones):
        yield [M.row(i - 1) for i in s]  # damn 1-indexing


def OneExactCover(M):
    """
    Use A. Ajanki's DLXMatrix class to solve the exact cover
    problem on the matrix M (treated as a dense binary matrix).

    EXAMPLES::

        sage: M = Matrix([[1,1,0],[1,0,1],[0,1,1]])  # no exact covers                  # needs sage.modules
        sage: OneExactCover(M)                                                          # needs sage.modules

        sage: M = Matrix([[1,1,0],[1,0,1],[0,0,1],[0,1,0]])  # two exact covers         # needs sage.modules
        sage: OneExactCover(M)                                                          # needs sage.modules
        [(1, 1, 0), (0, 0, 1)]
    """
    for s in AllExactCovers(M):
        return s



# sage_setup: distribution = sagemath-categories
r"""
Transitive ideal closure tool
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************


def transitive_ideal(f, x):
    r"""
    Return a list of all elements reachable from `x` in the abstract
    reduction system whose reduction relation is given by the function
    `f`.

    In more elementary terms:

    If `S` is a set, and `f` is a function sending every element of `S`
    to a list of elements of `S`, then we can define a digraph on the
    vertex set `S` by drawing an edge from `s` to `t` for every
    `s \in S` and every `t \in f(s)`.

    If `x \in S`, then an element `y \in S` is said to be reachable
    from `x` if there is a path `x \to y` in this graph.

    Given `f` and `x`, this method computes the list of all elements of
    `S` reachable from `x`.

    Note that if there are infinitely many such elements, then this
    method will never halt.

    For more powerful versions, see :mod:`sage.combinat.backtrack`

    EXAMPLES::

        sage: f = lambda x: [x-1] if x > 0 else []
        sage: sage.combinat.tools.transitive_ideal(f, 10)
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    known = [x]
    todo = [x]

    while todo:
        y = todo.pop(0)
        for z in f(y):
            if z in known:
                continue
            known.append(z)
            todo.append(z)

    known.sort()
    return known



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.rings.finite_rings
r"""
Algebraic closures of finite fields

Let `\Bold{F}` be a finite field, and let `\overline{\Bold{F}}` be an
algebraic closure of `\Bold{F}`; this is unique up to (non-canonical)
isomorphism.  For every `n\ge 1`, there is a unique subfield
`\Bold{F}_n` of `\overline{\Bold{F}}` such that
`\Bold{F}\subset\Bold{F}_n` and `[\Bold{F}_n:\Bold{F}]=n`.

In Sage, algebraic closures of finite fields are implemented using
compatible systems of finite fields.  The resulting Sage object keeps
track of a finite lattice of the subfields `\Bold{F}_n` and the
embeddings between them.  This lattice is extended as necessary.

The Sage class corresponding to `\overline{\Bold{F}}` can be
constructed from the finite field `\Bold{F}` by using the
:meth:`~sage.rings.finite_rings.finite_field_base.FiniteField.algebraic_closure`
method.

The Sage class for elements of `\overline{\Bold{F}}` is
:class:`AlgebraicClosureFiniteFieldElement`.  Such an element is
represented as an element of one of the `\Bold{F}_n`.  This means that
each element `x\in\Bold{F}` has infinitely many different
representations, one for each `n` such that `x` is in `\Bold{F}_n`.

.. NOTE::

    Only prime finite fields are currently accepted as base fields for
    algebraic closures.  To obtain an algebraic closure of a non-prime
    finite field `\Bold{F}`, take an algebraic closure of the prime
    field of `\Bold{F}` and embed `\Bold{F}` into this.

    Algebraic closures of finite fields are currently implemented
    using (pseudo-)Conway polynomials; see
    :class:`AlgebraicClosureFiniteField_pseudo_conway` and the module
    :mod:`~sage.rings.finite_rings.conway_polynomials`.  Other
    implementations may be added by creating appropriate subclasses of
    :class:`AlgebraicClosureFiniteField_generic`.

    In the current implementation, algebraic closures do not satisfy
    the unique parent condition.  Moreover, there is no coercion map
    between different algebraic closures of the same finite field.
    There is a conceptual reason for this, namely that the definition
    of pseudo-Conway polynomials only determines an algebraic closure
    up to *non-unique* isomorphism.  This means in particular that
    different algebraic closures, and their respective elements, never
    compare equal.

AUTHORS:

- Peter Bruin (August 2013): initial version

- Vincent Delecroix (November 2013): additional methods
"""

from sage.misc.abstract_method import abstract_method
from sage.misc.fast_methods import WithEqualityById

from sage.rings.finite_rings.finite_field_base import FiniteField
from sage.rings.ring import Field
from sage.structure.element import Element, FieldElement
from sage.structure.richcmp import richcmp


class AlgebraicClosureFiniteFieldElement(FieldElement):
    """
    Element of an algebraic closure of a finite field.

    EXAMPLES::

        sage: F = GF(3).algebraic_closure()
        sage: F.gen(2)
        z2
        sage: type(F.gen(2))
        <class 'sage.rings.algebraic_closure_finite_field.AlgebraicClosureFiniteField_pseudo_conway_with_category.element_class'>
    """
    def __init__(self, parent, value):
        """
        TESTS::

            sage: F = GF(3).algebraic_closure()
            sage: TestSuite(F.gen(2)).run(skip=['_test_pickling'])

        .. NOTE::

            The ``_test_pickling`` test has to be skipped because
            there is no coercion map between the parents of ``x``
            and ``loads(dumps(x))``.
        """
        if isinstance(value, Element) and isinstance(value.parent(), FiniteField):
            n = value.parent().degree()
        else:
            from sage.rings.integer import Integer
            n = Integer(1)
        self._value = parent._subfield(n).coerce(value)
        self._level = n
        FieldElement.__init__(self, parent)

    def __hash__(self):
        r"""
        TESTS::

            sage: F = GF(2).algebraic_closure()
            sage: hash(F.zero())
            0
            sage: hash(F.one())
            1
            sage: z1 = F.gen(1)
            sage: z2 = F.gen(2)
            sage: z3 = F.gen(3)
            sage: z4 = F.gen(4)

            sage: hash(z2) == hash(z3+z2-z3)
            True
            sage: hash(F.zero()) == hash(z3+z2-z3-z2)
            True

            sage: X = [z4**i for i in range(2**4-1)]
            sage: X.append(F.zero())
            sage: X.extend([z3, z3**2, z3*z4])
            sage: assert len(X) == len(set(hash(x) for x in X))

            sage: F = GF(3).algebraic_closure()
            sage: z1 = F.gen(1)
            sage: z2 = F.gen(2)
            sage: z3 = F.gen(3)
            sage: z4 = F.gen(4)

            sage: hash(z2) == hash(z3+z2-z3)
            True

            sage: X = [z4**i for i in range(3**4-1)]
            sage: X.append(F.zero())
            sage: X.extend([z3, z3**2, z3*z4])
            sage: assert len(X) == len(set(hash(x) for x in X))

        Check that :issue:`19956` is fixed::

            sage: R.<x,y> = GF(2).algebraic_closure()[]
            sage: x.resultant(y)
            y
        """
        # TODO: this is *very* slow
        # NOTE: the hash of a generator (e.g. z2, z3, ...) is always the
        # characteristic! In particular its hash value is not compatible with
        # sections.
        F, x, _ = self.as_finite_field_element(minimal=True)
        if F.degree() == 1:
            return hash(x)
        else:
            return hash((x, F.degree()))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F._repr_()
            'Algebraic closure of Finite Field of size 3'
        """
        return self._value._repr_()

    def _richcmp_(self, right, op):
        """
        Compare ``self`` with ``right``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2) == F.gen(3)
            False
        """
        x, y = self.parent()._to_common_subfield(self, right)
        return richcmp(x, y, op)

    def __pow__(self, exp):
        r"""
        TESTS::

            sage: F2 = GF(2).algebraic_closure()
            sage: z12 = F2.gen(3*4)
            sage: z12**3
            z12^3
            sage: z12**13
            z12^8 + z12^7 + z12^6 + z12^4 + z12^2 + z12
        """
        return self.__class__(self.parent(), self._value ** exp)

    def _add_(self, right):
        """
        Return ``self`` + ``right``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2) + F.gen(3)
            z6^5 + 2*z6^4 + 2*z6^3 + z6^2 + 2*z6 + 1
        """
        F = self.parent()
        x, y = F._to_common_subfield(self, right)
        return self.__class__(F, x + y)

    def _sub_(self, right):
        """
        Return ``self`` - ``right``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2) - F.gen(3)
            z6^4 + 2*z6^3 + z6^2 + 2*z6
        """
        F = self.parent()
        x, y = F._to_common_subfield(self, right)
        return self.__class__(F, x - y)

    def _mul_(self, right):
        """
        Return ``self`` * ``right``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2) * F.gen(3)
            z6^5 + 2*z6^4 + z6^2 + 2
        """
        F = self.parent()
        x, y = F._to_common_subfield(self, right)
        return self.__class__(F, x * y)

    def _div_(self, right):
        """
        Return ``self`` / ``right``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2) / F.gen(3)
            z6^5 + 2*z6^4 + z6^3 + 1
        """
        F = self.parent()
        x, y = F._to_common_subfield(self, right)
        return self.__class__(F, x / y)

    def change_level(self, n):
        """
        Return a representation of ``self`` as an element of the
        subfield of degree `n` of the parent, if possible.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: z = F.gen(4)
            sage: (z^10).change_level(6)
            2*z6^5 + 2*z6^3 + z6^2 + 2*z6 + 2
            sage: z.change_level(6)
            Traceback (most recent call last):
            ...
            ValueError: z4 is not in the image of Ring morphism:
              From: Finite Field in z2 of size 3^2
              To:   Finite Field in z4 of size 3^4
              Defn: z2 |--> 2*z4^3 + 2*z4^2 + 1

            sage: a = F(1).change_level(3); a
            1
            sage: a.change_level(2)
            1
            sage: F.gen(3).change_level(1)
            Traceback (most recent call last):
            ...
            ValueError: z3 is not in the image of Ring morphism:
              From: Finite Field of size 3
              To:   Finite Field in z3 of size 3^3
              Defn: 1 |--> 1
        """
        F = self.parent()
        l = self._level
        m = l.gcd(n)
        xl = self._value
        xm = F.inclusion(m, l).section()(xl)
        xn = F.inclusion(m, n)(xm)
        return self.__class__(F, xn)

    def _latex_(self):
        """
        Return a LaTeX representation of ``self``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: s = F.gen(1) + F.gen(2) + F.gen(3)
            sage: s
            z6^5 + 2*z6^4 + 2*z6^3 + z6^2 + 2*z6 + 2
            sage: latex(s)
            z_{6}^{5} + 2 z_{6}^{4} + 2 z_{6}^{3} + z_{6}^{2} + 2 z_{6} + 2
        """
        return self._value._latex_()

    def minpoly(self):
        """
        Return the minimal polynomial of ``self`` over the prime
        field.

        EXAMPLES::

            sage: F = GF(11).algebraic_closure()
            sage: F.gen(3).minpoly()
            x^3 + 2*x + 9
        """
        return self._value.minpoly()

    minimal_polynomial = minpoly

    def is_square(self):
        """
        Return ``True`` if ``self`` is a square.

        This always returns ``True``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2).is_square()
            True
        """
        return True

    def sqrt(self, all=False):
        """
        Return a square root of ``self``.

        If the optional keyword argument ``all`` is set to ``True``,
        return a list of all square roots of ``self`` instead.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.gen(2).sqrt()
            z4^3 + z4 + 1
            sage: F.gen(2).sqrt(all=True)
            [z4^3 + z4 + 1, 2*z4^3 + 2*z4 + 2]
            sage: (F.gen(2)^2).sqrt()
            z2
            sage: (F.gen(2)^2).sqrt(all=True)
            [z2, 2*z2]
        """
        F = self.parent()
        x = self._value
        if not x.is_square():
            l = self._level
            x = F.inclusion(l, 2*l)(x)
        sqrt = x.sqrt(extend=False, all=all)
        if all:
            return [self.__class__(F, y) for y in sqrt]
        return self.__class__(F, sqrt)

    def nth_root(self, n):
        """
        Return an `n`-th root of ``self``.

        EXAMPLES::

            sage: F = GF(5).algebraic_closure()
            sage: t = F.gen(2) + 1
            sage: s = t.nth_root(15); s
            4*z6^5 + 3*z6^4 + 2*z6^3 + 2*z6^2 + 4
            sage: s**15 == t
            True

        .. TODO::

            This function could probably be made faster.
        """
        from sage.rings.integer import Integer
        F = self.parent()
        x = self._value
        n = Integer(n)
        l = self._level
        # In order to be smart we look for the smallest subfield that
        # actually contains the root.
        for d in n.divisors():
            xx = F.inclusion(l, d*l)(x)
            try:
                y = xx.nth_root(n, extend=False)
            except ValueError:
                continue
            return self.__class__(F, y)

        raise AssertionError('cannot find n-th root in algebraic closure of finite field')

    def multiplicative_order(self):
        """
        Return the multiplicative order of ``self``.

        EXAMPLES::

            sage: K = GF(7).algebraic_closure()
            sage: K.gen(5).multiplicative_order()
            16806
            sage: (K.gen(1) + K.gen(2) + K.gen(3)).multiplicative_order()
            7353
        """
        return self._value.multiplicative_order()

    def pth_power(self, k=1):
        """
        Return the `p^k`-th power of ``self``, where `p` is the
        characteristic of ``self.parent()``.

        EXAMPLES::

            sage: K = GF(13).algebraic_closure('t')
            sage: t3 = K.gen(3)
            sage: s = 1 + t3 + t3**2
            sage: s.pth_power()
            10*t3^2 + 6*t3
            sage: s.pth_power(2)
            2*t3^2 + 6*t3 + 11
            sage: s.pth_power(3)
            t3^2 + t3 + 1
            sage: s.pth_power(3).parent() is K
            True
        """
        return self.__class__(self.parent(), self._value.pth_power(k))

    def pth_root(self, k=1):
        """
        Return the unique `p^k`-th root of ``self``, where `p` is the
        characteristic of ``self.parent()``.

        EXAMPLES::

            sage: K = GF(13).algebraic_closure('t')
            sage: t3 = K.gen(3)
            sage: s = 1 + t3 + t3**2
            sage: s.pth_root()
            2*t3^2 + 6*t3 + 11
            sage: s.pth_root(2)
            10*t3^2 + 6*t3
            sage: s.pth_root(3)
            t3^2 + t3 + 1
            sage: s.pth_root(2).parent() is K
            True
        """
        return self.__class__(self.parent(), self._value.pth_root(k))

    def as_finite_field_element(self, minimal=False):
        """
        Return ``self`` as a finite field element.

        INPUT:

        - ``minimal`` -- boolean (default: ``False``); if ``True``,
          always return the smallest subfield containing ``self``

        OUTPUT:

        - a triple (``field``, ``element``, ``morphism``) where
          ``field`` is a finite field, ``element`` an element of
          ``field`` and ``morphism`` a morphism from ``field`` to
          ``self.parent()``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure('t')
            sage: t = F.gen(5)
            sage: t.as_finite_field_element()
            (Finite Field in t5 of size 3^5,
             t5,
             Ring morphism:
              From: Finite Field in t5 of size 3^5
              To:   Algebraic closure of Finite Field of size 3
              Defn: t5 |--> t5)

        By default, ``field`` is not necessarily minimal.  We can
        force it to be minimal using the ``minimal`` option::

            sage: s = t + 1 - t
            sage: s.as_finite_field_element()[0]
            Finite Field in t5 of size 3^5
            sage: s.as_finite_field_element(minimal=True)[0]
            Finite Field of size 3

        This also works when the element has to be converted between
        two non-trivial finite subfields (see :issue:`16509`)::

            sage: K = GF(5).algebraic_closure()
            sage: z = K.gen(5) - K.gen(5) + K.gen(2)
            sage: z.as_finite_field_element(minimal=True)
            (Finite Field in z2 of size 5^2, z2, Ring morphism:
               From: Finite Field in z2 of size 5^2
               To:   Algebraic closure of Finite Field of size 5
               Defn: z2 |--> z2)

        There are automatic coercions between the various
        subfields::

            sage: a = K.gen(2) + 1
            sage: _,b,_ = a.as_finite_field_element()
            sage: K4 = K.subfield(4)[0]
            sage: K4(b)
            z4^3 + z4^2 + z4 + 4
            sage: b.minimal_polynomial() == K4(b).minimal_polynomial()
            True
            sage: K(K4(b)) == K(b)
            True

        You can also use the inclusions that are implemented at
        the level of the algebraic closure::

            sage: f = K.inclusion(2,4); f
            Ring morphism:
              From: Finite Field in z2 of size 5^2
              To:   Finite Field in z4 of size 5^4
              Defn: z2 |--> z4^3 + z4^2 + z4 + 3
            sage: f(b)
            z4^3 + z4^2 + z4 + 4
        """
        Fbar = self.parent()
        x = self._value
        l = self._level

        if minimal:
            m = x.minpoly().degree()
            if m == 1:
                x = Fbar.base_ring()(x)
            else:
                x = Fbar.inclusion(m, l).section()(x)
            l = m

        F, phi = Fbar.subfield(l)
        return (F, x, phi)


class AlgebraicClosureFiniteField_generic(Field):
    """
    Algebraic closure of a finite field.

    TESTS::

        sage: GF(3).algebraic_closure().cardinality()
        +Infinity

        sage: GF(3).algebraic_closure().is_finite()
        False
    """
    def __init__(self, base_ring, name, category=None):
        """
        TESTS::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField_generic
            sage: F = AlgebraicClosureFiniteField_generic(GF(5), 'z')
            sage: F
            Algebraic closure of Finite Field of size 5
        """
        Field.__init__(self, base_ring=base_ring, names=name,
                       normalize=False, category=category)

    def __eq__(self, other):
        """
        Compare ``self`` with ``other``.

        TESTS::

            sage: F3 = GF(3).algebraic_closure()
            sage: F3 == F3
            True
            sage: F5 = GF(5).algebraic_closure()
            sage: F3 == F5
            False
        """
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        return ((self.base_ring(), self.variable_name(), self.category()) ==
                (other.base_ring(), other.variable_name(), other.category()))

    def __ne__(self, other):
        """
        Check whether ``self`` and ``other`` are not equal.

        TESTS::

            sage: F3 = GF(3).algebraic_closure()
            sage: F3 != F3
            False
            sage: F5 = GF(5).algebraic_closure()
            sage: F3 != F5
            True
        """
        return not (self == other)

    def characteristic(self):
        """
        Return the characteristic of ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: p = next_prime(1000)
            sage: F = AlgebraicClosureFiniteField(GF(p), 'z')
            sage: F.characteristic() == p
            True
        """
        return self.base_ring().characteristic()

    Element = AlgebraicClosureFiniteFieldElement

    def _element_constructor_(self, x):
        """
        Construct an element of ``self``.

        TESTS::

            sage: F = GF(5).algebraic_closure()
            sage: type(F(3))
            <class 'sage.rings.algebraic_closure_finite_field.AlgebraicClosureFiniteField_pseudo_conway_with_category.element_class'>

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F1 = AlgebraicClosureFiniteField(GF(3), 'z')
            sage: F2 = AlgebraicClosureFiniteField(GF(3), 'z')
            sage: F1(F2.gen(1))
            Traceback (most recent call last):
            ...
            ValueError: no conversion defined between different algebraic closures
        """
        if isinstance(x, self.element_class):
            if x.parent() is not self:
                raise ValueError('no conversion defined between different algebraic closures')
            return x
        else:
            return self.element_class(self, x)

    def _coerce_map_from_(self, other):
        """
        Return ``True`` if elements of ``other`` can be coerced into
        ``self``.

        EXAMPLES::

            sage: F = GF(7).algebraic_closure()
            sage: F.has_coerce_map_from(Integers())
            True
        """
        if other is self:
            return True
        elif isinstance(other, FiniteField) and self._subfield(other.degree()) is other:
            return True
        elif self._subfield(1).has_coerce_map_from(other):
            return True

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'z')
            sage: F._repr_()
            'Algebraic closure of Finite Field of size 5'
        """
        return 'Algebraic closure of %s' % self.base_ring()

    def _latex_(self):
        r"""
        Return a LaTeX representation of ``self``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: latex(F)
            \overline{\Bold{F}_{3}}
        """
        return "\\overline{{{}}}".format(self.base_ring()._latex_())

    def _to_common_subfield(self, x, y):
        """
        Coerce `x` and `y` to a common subfield of ``self``.

        TESTS::

            sage: F = GF(3).algebraic_closure()
            sage: x, y = F._to_common_subfield(F.gen(2), F.gen(3))
            sage: x.parent()
            Finite Field in z6 of size 3^6
            sage: y.parent()
            Finite Field in z6 of size 3^6
        """
        if x._level == y._level:
            return x._value, y._value
        n = x._level.lcm(y._level)
        mx = self.inclusion(x._level, n)
        my = self.inclusion(y._level, n)
        return mx(x._value), my(y._value)

    @abstract_method
    def _get_polynomial(self, n):
        """
        Return the polynomial defining the unique subfield of degree
        `n` of ``self``.

        This must be implemented by subclasses.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField_generic
            sage: F = AlgebraicClosureFiniteField_generic(GF(5), 'z')
            sage: F._get_polynomial(1)
            Traceback (most recent call last):
            ...
            NotImplementedError: <abstract method _get_polynomial at ...>
        """

    @abstract_method
    def _get_im_gen(self, m, n):
        """
        Return the image of ``self.gen(m)`` under the canonical
        inclusion into ``self.subfield(n)``.

        This must be implemented by subclasses.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField_generic
            sage: F = AlgebraicClosureFiniteField_generic(GF(5), 'z')
            sage: F._get_im_gen(2, 4)
            Traceback (most recent call last):
            ...
            NotImplementedError: <abstract method _get_im_gen at ...>
        """

    def _subfield(self, n):
        """
        Return the unique subfield of degree `n` of ``self``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F._subfield(4)
            Finite Field in z4 of size 3^4
        """
        if n == 1:
            return self.base_ring()
        else:
            from sage.rings.finite_rings.finite_field_constructor import FiniteField
            return FiniteField(self.base_ring().cardinality() ** n,
                               name=self.variable_name() + str(n),
                               prefix=self.variable_name(),
                               modulus=self._get_polynomial(n),
                               check_irreducible=False)

    def subfield(self, n):
        """
        Return the unique subfield of degree `n` of ``self``
        together with its canonical embedding into ``self``.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.subfield(1)
            (Finite Field of size 3,
             Ring morphism:
               From: Finite Field of size 3
               To:   Algebraic closure of Finite Field of size 3
               Defn: 1 |--> 1)
            sage: F.subfield(4)
            (Finite Field in z4 of size 3^4,
             Ring morphism:
               From: Finite Field in z4 of size 3^4
               To:   Algebraic closure of Finite Field of size 3
               Defn: z4 |--> z4)
        """
        Fn = self._subfield(n)
        return Fn, Fn.hom( (self.gen(n),), check=False)

    def inclusion(self, m, n):
        """
        Return the canonical inclusion map from the subfield
        of degree `m` to the subfield of degree `n`.

        EXAMPLES::

            sage: F = GF(3).algebraic_closure()
            sage: F.inclusion(1, 2)
            Ring morphism:
              From: Finite Field of size 3
              To:   Finite Field in z2 of size 3^2
              Defn: 1 |--> 1
            sage: F.inclusion(2, 4)
            Ring morphism:
              From: Finite Field in z2 of size 3^2
              To:   Finite Field in z4 of size 3^4
              Defn: z2 |--> 2*z4^3 + 2*z4^2 + 1
        """
        if m.divides(n):
            # check=False is required to avoid "coercion hell": an
            # infinite loop in checking the morphism involving
            # polynomial_compiled.pyx on the modulus().
            return self._subfield(m).hom( (self._get_im_gen(m, n),), check=False)
        else:
            raise ValueError("subfield of degree %s not contained in subfield of degree %s" % (m, n))

    def ngens(self):
        """
        Return the number of generators of ``self``, which is
        infinity.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: AlgebraicClosureFiniteField(GF(5), 'z').ngens()
            +Infinity
        """
        from sage.rings.infinity import Infinity
        return Infinity

    def gen(self, n):
        """
        Return the `n`-th generator of ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'z')
            sage: F.gen(2)
            z2
        """
        F = self._subfield(n)
        return self(F.gen())

    def gens(self):
        """
        Return a family of generators of ``self``.

        OUTPUT:

        - a :class:`~sage.sets.family.Family`, indexed by the positive
          integers, whose `n`-th element is ``self.gen(n)``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'z')
            sage: g = F.gens(); g
            Lazy family (...(i))_{i in Positive integers}
            sage: g[3]
            z3
        """
        from sage.sets.family import Family
        from sage.sets.positive_integers import PositiveIntegers
        return Family(PositiveIntegers(), self.gen)

    def _first_ngens(self, n):
        """
        Return the first `n` generators of ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'z')
            sage: F._first_ngens(3)
            (1, z2, z3)
        """
        return tuple(self.gen(i + 1) for i in range(n))

    def algebraic_closure(self):
        """
        Return an algebraic closure of ``self``.

        This always returns ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'z')
            sage: F.algebraic_closure() is F
            True
        """
        return self

    def _an_element_(self):
        """
        TESTS::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
            sage: F = AlgebraicClosureFiniteField(GF(5), 'w')
            sage: F.an_element()  # indirect doctest
            w2
        """
        return self.gen(2)

    def some_elements(self):
        r"""
        Return some elements of this field.

        EXAMPLES::

            sage: F = GF(7).algebraic_closure()
            sage: F.some_elements()
            (1, z2, z3 + 1)
        """
        return (self(1), self.gen(2), 1+self.gen(3))

    def _roots_univariate_polynomial(self, p, ring=None, multiplicities=None, algorithm=None):
        r"""
        Return a list of pairs ``(root,multiplicity)`` of roots of the polynomial ``p``.

        If the argument ``multiplicities`` is set to ``False`` then return the
        list of roots.

        .. SEEALSO::

            :meth:`_factor_univariate_polynomial`

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(5),'x')
            sage: K = GF(5).algebraic_closure('t')

            sage: sorted((x^6 - 1).roots(K,multiplicities=False))
            [1, 4, 2*t2 + 1, 2*t2 + 2, 3*t2 + 3, 3*t2 + 4]
            sage: ((K.gen(2)*x - K.gen(3))**2).roots(K)
            [(3*t6^5 + 2*t6^4 + 2*t6^2 + 3, 2)]

            sage: for _ in range(10):
            ....:     p = R.random_element(degree=randint(2,8))
            ....:     for r in p.roots(K, multiplicities=False):
            ....:         assert p(r).is_zero(), "r={} is not a root of p={}".format(r,p)
        """
        from sage.arith.functions import lcm
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

        # first build a polynomial over some finite field
        coeffs = [v.as_finite_field_element(minimal=True) for v in p.list()]
        l = lcm([c[0].degree() for c in coeffs])
        F, phi = self.subfield(l)
        P = p.parent().change_ring(F)

        new_coeffs = [self.inclusion(c[0].degree(), l)(c[1]) for c in coeffs]

        polys = [(g,m,l,phi) for g,m in P(new_coeffs).factor()]
        roots = []    # a list of pair (root,multiplicity)
        while polys:
            g,m,l,phi = polys.pop()

            if g.degree() == 1: # found a root
                r = phi(-g.constant_coefficient())
                roots.append((r,m))
            else: # look at the extension of degree g.degree() which contains at
                  # least one root of g
                ll = l * g.degree()
                psi = self.inclusion(l, ll)
                FF, pphi = self.subfield(ll)
                # note: there is no coercion from the l-th subfield to the ll-th
                # subfield. The line below does the conversion manually.
                g = PolynomialRing(FF, 'x')([psi(_) for _ in g])
                polys.extend((gg,m,ll,pphi) for gg,_ in g.factor())

        if multiplicities:
            return roots
        else:
            return [r[0] for r in roots]

    def _factor_univariate_polynomial(self, p, **kwds):
        r"""
        Factorization of univariate polynomials.

        EXAMPLES::

            sage: K = GF(3).algebraic_closure()
            sage: R = PolynomialRing(K, 'T')
            sage: T = R.gen()
            sage: (K.gen(2) * T^2 - 1).factor()
            (z2) * (T + z4^3 + z4^2 + z4) * (T + 2*z4^3 + 2*z4^2 + 2*z4)

            sage: for d in range(10):
            ....:     p = R.random_element(degree=randint(2,8))
            ....:     assert p.factor().prod() == p, "error in the factorization of p={}".format(p)
        """
        from sage.structure.factorization import Factorization
        R = p.parent()
        return Factorization([(R([-root, self.one()]), m) for root, m in p.roots()], unit=p[p.degree()])


class AlgebraicClosureFiniteField_pseudo_conway(WithEqualityById, AlgebraicClosureFiniteField_generic):
    """
    Algebraic closure of a finite field, constructed using
    pseudo-Conway polynomials.

    EXAMPLES::

        sage: F = GF(5).algebraic_closure(implementation='pseudo_conway')
        sage: F.cardinality()
        +Infinity
        sage: F.algebraic_closure() is F
        True
        sage: x = F(3).nth_root(12); x
        z4^3 + z4^2 + 4*z4
        sage: x**12
        3

    TESTS::

        sage: F3 = GF(3).algebraic_closure()
        sage: F3 == F3
        True
        sage: F5 = GF(5).algebraic_closure()
        sage: F3 == F5
        False
    """
    def __init__(self, base_ring, name, category=None, lattice=None, use_database=True):
        """
        INPUT:

        - ``base_ring`` -- the finite field of which to construct an
          algebraic closure.  Currently only prime fields are
          accepted.

        - ``name`` -- prefix to use for generators of the finite
          subfields

        - ``category`` -- if provided, specifies the category in which
          this algebraic closure will be placed

        - ``lattice`` -- :class:`~sage.rings.finite_rings.conway_polynomials.PseudoConwayPolynomialLattice`
          (default: ``None``); if provided, use this pseudo-Conway
          polynomial lattice to construct an algebraic closure.

        - ``use_database`` -- boolean.  If ``True`` (default), use actual
          Conway polynomials whenever they are available in the
          database.  If ``False``, always compute pseudo-Conway
          polynomials from scratch.

        TESTS::

            sage: F = GF(5).algebraic_closure(implementation='pseudo_conway')
            sage: print(F.__class__.__name__)
            AlgebraicClosureFiniteField_pseudo_conway_with_category
            sage: TestSuite(F).run(skip=['_test_elements', '_test_pickling'])

            sage: from sage.rings.finite_rings.conway_polynomials import PseudoConwayLattice
            sage: L = PseudoConwayLattice(11, use_database=False)
            sage: F = GF(7).algebraic_closure(lattice=L)
            Traceback (most recent call last):
            ...
            TypeError: lattice must be a pseudo-Conway lattice with characteristic 7
            sage: F = GF(11).algebraic_closure(lattice=L)
            sage: F.gen(2).minimal_polynomial()
            x^2 + 4*x + 2

            sage: F = GF(11).algebraic_closure(use_database=True)
            sage: F.gen(2).minimal_polynomial()
            x^2 + 7*x + 2

        .. NOTE::

            In the test suite, ``_test_pickling`` has to be skipped
            because ``F`` and ``loads(dumps(F))`` cannot consistently
            be made to compare equal, and ``_test_elements`` has to be
            skipped for the reason described in
            :meth:`AlgebraicClosureFiniteFieldElement.__init__`.
        """
        if not (isinstance(base_ring, FiniteField) and base_ring.is_prime_field()):
            raise NotImplementedError('algebraic closures of finite fields are only implemented for prime fields')
        from sage.rings.finite_rings.conway_polynomials import PseudoConwayLattice
        p = base_ring.characteristic()
        if lattice is None:
            lattice = PseudoConwayLattice(p, use_database)
        elif not isinstance(lattice, PseudoConwayLattice) or lattice.p != p:
            raise TypeError('lattice must be a pseudo-Conway lattice with characteristic %s' % p)
        self._pseudo_conway_lattice = lattice
        AlgebraicClosureFiniteField_generic.__init__(self, base_ring, name, category)

    def _get_polynomial(self, n):
        """
        Return the defining polynomial of the unique subfield of
        degree `n` of ``self``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField_pseudo_conway
            sage: F = AlgebraicClosureFiniteField_pseudo_conway(GF(5), 'z')
            sage: F._get_polynomial(1)
            x + 3
        """
        return self._pseudo_conway_lattice.polynomial(n)

    def _get_im_gen(self, m, n):
        """
        Return the image of ``self.gen(m)`` under the canonical
        inclusion into ``self.subfield(n)``.

        EXAMPLES::

            sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField_pseudo_conway
            sage: F = AlgebraicClosureFiniteField_pseudo_conway(GF(5), 'z')
            sage: F._get_im_gen(2, 4)
            z4^3 + z4^2 + z4 + 3
        """
        p = self.characteristic()
        if m == 1:
            return self._subfield(n).one()
        return self._subfield(n).gen() ** ((p**n - 1)//(p**m - 1))


def AlgebraicClosureFiniteField(base_ring, name, category=None, implementation=None, **kwds):
    """
    Construct an algebraic closure of a finite field.

    The recommended way to use this functionality is by calling the
    :meth:`~sage.rings.finite_rings.finite_field_base.FiniteField.algebraic_closure`
    method of the finite field.

    .. NOTE::

        Algebraic closures of finite fields in Sage do not have the
        unique representation property, because they are not
        determined up to unique isomorphism by their defining data.

    EXAMPLES::

        sage: from sage.rings.algebraic_closure_finite_field import AlgebraicClosureFiniteField
        sage: F = GF(2).algebraic_closure()
        sage: F1 = AlgebraicClosureFiniteField(GF(2), 'z')
        sage: F1 is F
        False

    In the pseudo-Conway implementation, non-identical instances never
    compare equal::

        sage: F1 == F
        False
        sage: loads(dumps(F)) == F
        False

    This is to ensure that the result of comparing two instances
    cannot change with time.
    """
    if category is None:
        from sage.categories.fields import Fields
        category = Fields().Infinite()

    if implementation is None:
        implementation = 'pseudo_conway'

    if implementation == 'pseudo_conway':
        return AlgebraicClosureFiniteField_pseudo_conway(base_ring, name, category, **kwds)
    else:
        raise ValueError('unknown implementation for algebraic closure of finite field: %s'
                         % implementation)



# sage_setup: distribution = sagemath-categories
"""
Binary trees

This implements a binary tree in Cython.

AUTHORS:

- Tom Boothby (2007-02-15).  Initial version free for any use (public domain).
"""
from cysignals.memory cimport sig_malloc, sig_free

from cpython.ref cimport PyObject, Py_INCREF, Py_XDECREF

cdef binary_tree_node *BinaryTreeNode(int key, object value) noexcept:
    cdef binary_tree_node *t
    t = <binary_tree_node *>sig_malloc(sizeof(binary_tree_node))
    t.key = key
    t.left = NULL
    t.right = NULL
    Py_INCREF(value)
    t.value = <void *>value
    return t

cdef void free_binary_tree_node(binary_tree_node *self) noexcept:
    Py_XDECREF(<PyObject *>self.value)
    sig_free(self)

cdef inline void binary_tree_dealloc(binary_tree_node *self) noexcept:
    if self != NULL:
        binary_tree_dealloc(self.left)
        binary_tree_dealloc(self.right)
        free_binary_tree_node(self)


cdef void binary_tree_insert(binary_tree_node *self, int key, object value) noexcept:
    if self.key == key:
        return
    elif self.key > key:
        if self.left == NULL:
            self.left = BinaryTreeNode(key, value)
        else:
            binary_tree_insert(self.left, key, value)
    else:
        if self.right == NULL:
            self.right = BinaryTreeNode(key, value)
        else:
            binary_tree_insert(self.right, key, value)

cdef object binary_tree_get(binary_tree_node *self, int key):
    if self.key == key:
        return <object>self.value
    elif self.key > key:
        if self.left == NULL:
            return None
        else:
            return binary_tree_get(self.left, key)
    else:
        if self.right == NULL:
            return None
        else:
            return binary_tree_get(self.right, key)

cdef object binary_tree_delete(binary_tree_node *self, int key):
    cdef object t
    if self.key > key:
        if self.left == NULL:
            return None
        elif self.left.key == key:
            t = <object>self.left.value
            self.left = binary_tree_left_excise(self.left)
            return t
        else:
            return binary_tree_delete(self.left, key)
    else:
        if self.right == NULL:
            return None
        elif self.right.key == key:
            t = <object>self.right.value
            self.right = binary_tree_right_excise(self.right)
            return t
        else:
            return binary_tree_delete(self.right, key)

cdef binary_tree_node *binary_tree_left_excise(binary_tree_node *self) noexcept:
    cdef binary_tree_node *left
    cdef binary_tree_node *cur
    if self.left == NULL:
        left = self.right
    elif self.right == NULL:
        left = self.left
    else:
        left = self.left
        cur = self.right
        while cur.right != NULL:
            cur = cur.right
        cur.right = self.left.right
    free_binary_tree_node(self)
    return left


cdef binary_tree_node *binary_tree_right_excise(binary_tree_node *self) noexcept:
    cdef binary_tree_node *right
    cdef binary_tree_node *cur
    if self.right == NULL:
        right = self.left
    elif self.left == NULL:
        right = self.right
    else:
        right = self.right
        cur = self.left
        while cur.left != NULL:
            cur = cur.left
        cur.left = self.right.left
    free_binary_tree_node(self)
    return right


cdef binary_tree_node *binary_tree_head_excise(binary_tree_node *self) noexcept:
    cdef binary_tree_node *cur
    cdef int right
    # We have a pointer we're about to free.  Chances are, we'll never
    # see this pointer again.  Thus, its least significant bit is
    # "random" enough to resist bias.
    right = (<int>self)&1
    if self.right == NULL:
        return self.left
    if self.left == NULL:
        return self.right
    if right:
        # move right branch to left, return left
        cur = self.left
        while cur.right != NULL:
            cur = cur.right
        cur.right = self.right
        cur = self.left
    else:
        # move left branch to right, return right
        cur = self.right
        while cur.left != NULL:
            cur = cur.left
        cur.left = self.left
        cur = self.right
    free_binary_tree_node(self)
    return cur


cdef int LIST_PREORDER, LIST_POSTORDER, LIST_INORDER, LIST_KEYS, LIST_VALUES
LIST_PREORDER  = 1
LIST_INORDER = 2
LIST_POSTORDER = 4
LIST_KEYS = 8
LIST_VALUES = 16

cdef object binary_tree_list(binary_tree_node *cur, int behavior):
    if behavior & LIST_KEYS:
        item = int(cur.key)
    else:
        item = <object>cur.value

    if behavior & LIST_PREORDER:
        arry = [item]
    else:
        arry = []

    if cur.left != NULL:
        arry.extend(binary_tree_list(cur.left, behavior))

    if behavior & LIST_INORDER:
        arry.append(item)

    if cur.right != NULL:
        arry.extend(binary_tree_list(cur.right, behavior))

    if behavior & LIST_POSTORDER:
        arry.append(item)

    return arry


cdef class BinaryTree:
    """
    A simple binary tree with integer keys.
    """
    def __cinit__(BinaryTree self):
        self.head = NULL

    def __dealloc__(BinaryTree self):
        """
        TESTS:

        We test that :issue:`18897` is fixed::

            sage: def test():
            ....:     from sage.rings.polynomial.polynomial_compiled import CompiledPolynomialFunction
            ....:     import gc
            ....:     from collections import Counter
            ....:     gc.collect()
            ....:     pre={id(c) for c in gc.get_objects()}
            ....:     L = [-1, 9, -22, 21, -8, 1]
            ....:     for _ in range(100):
            ....:         CompiledPolynomialFunction(L)  # this creates and deallocs a binary tree
            ....:     gc.collect()
            ....:     post=Counter(type(o) for o in gc.get_objects() if id(o) not in pre)
            ....:     return [(k,v) for (k,v) in post.items() if v>10]
            sage: test()   # indirect doctest
            []
        """
        binary_tree_dealloc(self.head)

    def insert(BinaryTree self, object key, object value=None):
        """
        Insert a key-value pair into the BinaryTree.

        Duplicate keys are ignored.

        The first parameter, key, should be an int, or coercible (one-to-one)
        into an int.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.insert(1)
            sage: t.insert(0)
            sage: t.insert(2)
            sage: t.insert(0,1)
            sage: t.get(0)
            0
        """
        cdef int ckey
        if value is None:
            value = key
        ckey = int(key)
        if self.head is NULL:
            self.head = BinaryTreeNode(ckey, value)
        else:
            binary_tree_insert(self.head, ckey, value)

    def delete(BinaryTree self, int key):
        """
        Remove a the node corresponding to key, and return the value
        associated with it.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.insert(3,3)
            sage: t.insert(1,1)
            sage: t.insert(2,2)
            sage: t.insert(0,0)
            sage: t.insert(5,5)
            sage: t.insert(6,6)
            sage: t.insert(4,4)
            sage: t.delete(0)
            0
            sage: t.delete(3)
            3
            sage: t.delete(5)
            5
            sage: t.delete(2)
            2
            sage: t.delete(6)
            6
            sage: t.delete(1)
            1
            sage: t.delete(0)
            sage: t.get_max()
            4
            sage: t.get_min()
            4
        """
        cdef object r
        if self.head == NULL:
            return None
        elif self.head.key == key:
            r = <object>self.head.value
            self.head = binary_tree_head_excise(self.head)
            return r
        else:
            return binary_tree_delete(self.head, key)

    def get(BinaryTree self, int key):
        """
        Return the value associated with the key given.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.insert(0, Matrix([[0,0], [1,1]]))                                   # needs sage.modules
            sage: t.insert(0, 1)
            sage: t.get(0)                                                              # needs sage.modules
            [0 0]
            [1 1]
        """
        if self.head == NULL:
            return None
        return binary_tree_get(self.head, key)

    def contains(BinaryTree self, int key):
        """
        Return whether a node with the given key exists in the tree.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.contains(1)
            False
            sage: t.insert(1, 1)
            sage: t.contains(1)
            True
        """
        if self.head == NULL:
            return False
        return binary_tree_get(self.head, key) is not None

    def get_max(BinaryTree self):
        """
        Return the value of the node with the maximal key value.
        """
        cdef binary_tree_node *cur
        if self.head == NULL:
            return None
        cur = self.head
        while cur.right != NULL:
            cur = cur.right
        return <object>cur.value

    def get_min(BinaryTree self):
        """
        Return the value of the node with the minimal key value.
        """
        cdef binary_tree_node *cur
        if self.head == NULL:
            return None
        cur = self.head
        while cur.left != NULL:
            cur = cur.left
        return <object>cur.value

    def pop_max(BinaryTree self):
        """
        Return the value of the node with the maximal key value,
        and remove that node from the tree.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.insert(4,'e')
            sage: t.insert(2,'c')
            sage: t.insert(0,'a')
            sage: t.insert(1,'b')
            sage: t.insert(3,'d')
            sage: t.insert(5,'f')
            sage: while not t.is_empty():
            ....:     print(t.pop_max())
            f
            e
            d
            c
            b
            a
        """
        cdef binary_tree_node *cur
        cdef object max
        if self.head == NULL:
            return None
        if self.head.right == NULL:
            max = <object>self.head.value
            cur = self.head.left
            free_binary_tree_node(self.head)
            self.head = cur
            return max
        cur = self.head
        while cur.right.right != NULL:
            cur = cur.right
        max = <object>cur.right.value
        cur.right = binary_tree_right_excise(cur.right)
        return max

    def pop_min(BinaryTree self):
        """
        Return the value of the node with the minimal key value,
        and remove that node from the tree.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.insert(4,'e')
            sage: t.insert(2,'c')
            sage: t.insert(0,'a')
            sage: t.insert(1,'b')
            sage: t.insert(3,'d')
            sage: t.insert(5,'f')
            sage: while not t.is_empty():
            ....:     print(t.pop_min())
            a
            b
            c
            d
            e
            f
        """
        cdef binary_tree_node *cur
        cdef object min
        if self.head == NULL:
            return None
        if self.head.left == NULL:
            min = <object>self.head.value
            cur = self.head.right
            free_binary_tree_node(self.head)
            self.head = cur
            return min
        cur = self.head
        while cur.left.left != NULL:
            cur = cur.left
        min = <object>cur.left.value
        cur.left = binary_tree_left_excise(cur.left)
        return min

    def is_empty(BinaryTree self):
        """
        Return whether the tree has no nodes.

        EXAMPLES::

            sage: from sage.misc.binary_tree import BinaryTree
            sage: t = BinaryTree()
            sage: t.is_empty()
            True
            sage: t.insert(0,0)
            sage: t.is_empty()
            False
        """
        return self.head == NULL

    def keys(BinaryTree self, order='inorder'):
        """
        Return the keys sorted according to "order" parameter.

        The order can be one of
        "inorder", "preorder", or "postorder"
        """
        if self.head == NULL:
            return []

        if order == "postorder":
            o = LIST_POSTORDER
        elif order == "inorder":
            o = LIST_INORDER
        else:
            o = LIST_PREORDER

        return binary_tree_list(self.head, LIST_KEYS + o)

    def values(BinaryTree self, order='inorder'):
        """
        Return the keys sorted according to "order" parameter.

        The order can be one of
        "inorder", "preorder", or "postorder"
        """
        if self.head == NULL:
            return []

        if order == "postorder":
            o = LIST_POSTORDER
        elif order == "inorder":
            o = LIST_INORDER
        else:
            o = LIST_PREORDER

        return binary_tree_list(self.head, LIST_VALUES + o)

    def _headkey_(BinaryTree self):
        """
        Used by the stress tester.  Don't think a user would care.

        Email tom if you care what the headkey is.
        """
        if self.head == NULL:
            return 0
        return self.head.key


class Test:
    def random(self):
        self.binary_tree()

    def binary_tree(self, values=100, cycles=100000):
        """
        Perform a sequence of random operations, given random inputs
        to stress test the binary tree structure.

        This was useful during development to find memory leaks /
        segfaults.  Cycles should be at least 100 times as large as
        values, or the delete, contains, and get methods won't hit
        very often.

        INPUT:

        - ``values`` -- number of possible values to use

        - ``cycles`` -- number of operations to perform

        TESTS::

            sage: sage.misc.binary_tree.Test().random()
        """
        from sage.misc.prandom import randint
        t = BinaryTree()
        for i in range(cycles):
            r = randint(0, 8)
            s = randint(0, values)
            if r == 1:
                t.insert(s)
            elif r == 2:
                t.delete(t._headkey_())
            elif r == 3:
                t.get(s)
            elif r == 4:
                t.contains(s)
            elif r == 5:
                t.get_max()
            elif r == 6:
                t.get_min()
            elif r == 7:
                t.pop_min()
            elif r == 8:
                t.pop_max()
            else:
                t.delete(s)



# sage_setup: distribution = sagemath-categories
r"""
Generic data structures and algorithms for rings

AUTHORS:

- Lorenz Panny (2022): :class:`ProductTree`, :func:`prod_with_derivative`
"""

from sage.misc.misc_c import prod


class ProductTree:
    r"""
    A simple binary product tree, i.e., a tree of ring elements in
    which every node equals the product of its children.
    (In particular, the *root* equals the product of all *leaves*.)

    Product trees are a very useful building block for fast computer
    algebra. For example, a quasilinear-time Discrete Fourier Transform
    (the famous *Fast* Fourier Transform) can be implemented as follows
    using the :meth:`remainders` method of this class::

        sage: # needs sage.rings.finite_rings
        sage: from sage.rings.generic import ProductTree
        sage: F = GF(65537)
        sage: a = F(1111)
        sage: assert a.multiplicative_order() == 1024
        sage: R.<x> = F[]
        sage: ms = [x - a^i for i in range(1024)]               # roots of unity
        sage: ys = [F.random_element() for _ in range(1024)]    # input vector
        sage: tree = ProductTree(ms)
        sage: zs = tree.remainders(R(ys))                       # compute FFT!
        sage: zs == [R(ys) % m for m in ms]
        True

    Similarly, the :meth:`interpolation` method can be used to implement
    the inverse Fast Fourier Transform::

        sage: tree.interpolation(zs).padded_list(len(ys)) == ys                         # needs sage.rings.finite_rings
        True

    This class encodes the tree as *layers*: Layer `0` is just a tuple
    of the leaves. Layer `i+1` is obtained from layer `i` by replacing
    each pair of two adjacent elements by their product, starting from
    the left. (If the length is odd, the unpaired element at the end is
    simply copied as is.) This iteration stops as soon as it yields a
    layer containing only a single element (the root).

    .. NOTE::

        Use this class if you need the :meth:`remainders` method.
        To compute just the product, :func:`prod` is likely faster.

    INPUT:

    - ``leaves`` -- an iterable of elements in a common ring

    EXAMPLES::

        sage: from sage.rings.generic import ProductTree
        sage: R.<x> = GF(101)[]
        sage: vs = [x - i for i in range(1,10)]
        sage: tree = ProductTree(vs)
        sage: tree.root()
        x^9 + 56*x^8 + 62*x^7 + 44*x^6 + 47*x^5 + 42*x^4 + 15*x^3 + 11*x^2 + 12*x + 13
        sage: tree.remainders(x^7 + x + 1)
        [3, 30, 70, 27, 58, 72, 98, 98, 23]
        sage: tree.remainders(x^100)
        [1, 1, 1, 1, 1, 1, 1, 1, 1]

    ::

        sage: # needs sage.libs.pari
        sage: vs = prime_range(100)
        sage: tree = ProductTree(vs)
        sage: tree.root().factor()
        2 * 3 * 5 * 7 * 11 * 13 * 17 * 19 * 23 * 29 * 31 * 37 * 41 * 43 * 47 * 53 * 59 * 61 * 67 * 71 * 73 * 79 * 83 * 89 * 97
        sage: tree.remainders(3599)
        [1, 2, 4, 1, 2, 11, 12, 8, 11, 3, 3, 10, 32, 30, 27, 48, 0, 0, 48, 49, 22, 44, 30, 39, 10]

    We can access the individual layers of the tree::

        sage: tree.layers                                                               # needs sage.libs.pari
        [(2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97),
         (6, 35, 143, 323, 667, 1147, 1763, 2491, 3599, 4757, 5767, 7387, 97),
         (210, 46189, 765049, 4391633, 17120443, 42600829, 97),
         (9699690, 3359814435017, 729345064647247, 97),
         (32589158477190044730, 70746471270782959),
         (2305567963945518424753102147331756070,)]
    """
    def __init__(self, leaves):
        r"""
        Initialize a product tree having the given ring elements
        as its leaves.

        EXAMPLES::

            sage: from sage.rings.generic import ProductTree
            sage: vs = prime_range(100)                                                 # needs sage.libs.pari
            sage: tree = ProductTree(vs)                                                # needs sage.libs.pari
        """
        V = tuple(leaves)
        self.layers = [V]
        while len(V) > 1:
            V = tuple(prod(V[i:i+2]) for i in range(0,len(V),2))
            self.layers.append(V)

    def __len__(self):
        r"""
        Return the number of leaves of this product tree.

        EXAMPLES::

            sage: from sage.rings.generic import ProductTree
            sage: R.<x> = GF(101)[]
            sage: vs = [x - i for i in range(1,10)]
            sage: tree = ProductTree(vs)
            sage: len(tree)
            9
            sage: len(tree) == len(vs)
            True
            sage: len(tree.remainders(x^2))
            9
        """
        return len(self.layers[0])

    def __iter__(self):
        r"""
        Return an iterator over the leaves of this product tree.

        EXAMPLES::

            sage: from sage.rings.generic import ProductTree
            sage: R.<x> = GF(101)[]
            sage: vs = [x - i for i in range(1,10)]
            sage: tree = ProductTree(vs)
            sage: next(iter(tree)) == vs[0]
            True
            sage: list(tree) == vs
            True
        """
        return iter(self.layers[0])

    def root(self):
        r"""
        Return the value at the root of this product tree (i.e., the product of all leaves).

        EXAMPLES::

            sage: from sage.rings.generic import ProductTree
            sage: R.<x> = GF(101)[]
            sage: vs = [x - i for i in range(1,10)]
            sage: tree = ProductTree(vs)
            sage: tree.root()
            x^9 + 56*x^8 + 62*x^7 + 44*x^6 + 47*x^5 + 42*x^4 + 15*x^3 + 11*x^2 + 12*x + 13
            sage: tree.root() == prod(vs)
            True
        """
        assert len(self.layers[-1]) == 1
        return self.layers[-1][0]

    def leaves(self):
        r"""
        Return a tuple containing the leaves of this product tree.

        EXAMPLES::

            sage: from sage.rings.generic import ProductTree
            sage: R.<x> = GF(101)[]
            sage: vs = [x - i for i in range(1,10)]
            sage: tree = ProductTree(vs)
            sage: tree.leaves()
            (x + 100, x + 99, x + 98, ..., x + 93, x + 92)
            sage: tree.leaves() == tuple(vs)
            True
        """
        return self.layers[0]

    def remainders(self, x):
        r"""
        Given a value `x`, return a list of all remainders of `x`
        modulo the leaves of this product tree.

        The base ring must support the ``%`` operator for this
        method to work.

        INPUT:

        - ``x`` -- an element of the base ring of this product tree

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: from sage.rings.generic import ProductTree
            sage: vs = prime_range(100)
            sage: tree = ProductTree(vs)
            sage: n = 1085749272377676749812331719267
            sage: tree.remainders(n)
            [1, 1, 2, 1, 9, 1, 7, 15, 8, 20, 15, 6, 27, 11, 2, 6, 0, 25, 49, 5, 51, 4, 19, 74, 13]
            sage: [n % v for v in vs]
            [1, 1, 2, 1, 9, 1, 7, 15, 8, 20, 15, 6, 27, 11, 2, 6, 0, 25, 49, 5, 51, 4, 19, 74, 13]
        """
        X = [x]
        for V in reversed(self.layers):
            X = [X[i // 2] % V[i] for i in range(len(V))]
        return X

    _crt_bases = None

    def interpolation(self, xs):
        r"""
        Given a sequence ``xs`` of values, one per leaf, return a
        single element `x` which is congruent to the `i`\th value in
        ``xs`` modulo the `i`\th leaf, for all `i`.

        This is an explicit version of the Chinese remainder theorem;
        see also :meth:`CRT`. Using this product tree is faster for
        repeated calls since the required CRT bases are cached after
        the first run.

        The base ring must support the :func:`xgcd` function for this
        method to work.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: from sage.rings.generic import ProductTree
            sage: vs = prime_range(100)
            sage: tree = ProductTree(vs)
            sage: tree.interpolation([1, 1, 2, 1, 9, 1, 7, 15, 8, 20, 15, 6, 27, 11, 2, 6, 0, 25, 49, 5, 51, 4, 19, 74, 13])
            1085749272377676749812331719267

        This method is faster than :func:`CRT` for repeated calls with
        the same moduli::

            sage: # needs sage.libs.pari
            sage: vs = prime_range(1000,2000)
            sage: rs = lambda: [randrange(1,100) for _ in vs]
            sage: tree = ProductTree(vs)
            sage: %timeit CRT(rs(), vs)             # not tested
            372 µs ± 3.34 µs per loop (mean ± std. dev. of 7 runs, 1,000 loops each)
            sage: %timeit tree.interpolation(rs())  # not tested
            146 µs ± 479 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
        """
        if self._crt_bases is None:
            from sage.arith.misc import CRT_basis
            self._crt_bases = []
            for V in self.layers[:-1]:
                B = tuple(CRT_basis(V[i:i+2]) for i in range(0, len(V), 2))
                self._crt_bases.append(B)
        if len(xs) != len(self.layers[0]):
            raise ValueError('number of given elements must equal the number of leaves')
        for basis, layer in zip(self._crt_bases, self.layers[1:]):
            xs = [sum(c*x for c, x in zip(cs, xs[2*i:2*i+2])) % mod
                  for i, (cs, mod) in enumerate(zip(basis, layer))]
        assert len(xs) == 1
        return xs[0]


def prod_with_derivative(pairs):
    r"""
    Given an iterable of pairs `(f, \partial f)` of ring elements,
    return the pair `(\prod f, \partial \prod f)`, assuming `\partial`
    is an operator obeying the standard product rule.

    This function is entirely algebraic, hence still works when the
    elements `f` and `\partial f` are all passed through some ring
    homomorphism first. One particularly useful instance of this is
    evaluating the derivative of a product of polynomials at a point
    without fully expanding the product; see the second example below.

    INPUT:

    - ``pairs`` -- an iterable of tuples `(f, \partial f)` of elements
      of a common ring

    ALGORITHM: Repeated application of the product rule.

    EXAMPLES::

        sage: from sage.rings.generic import prod_with_derivative
        sage: R.<x> = ZZ[]
        sage: fs = [x^2 + 2*x + 3, 4*x + 5, 6*x^7 + 8*x + 9]
        sage: prod(fs)
        24*x^10 + 78*x^9 + 132*x^8 + 90*x^7 + 32*x^4 + 140*x^3 + 293*x^2 + 318*x + 135
        sage: prod(fs).derivative()
        240*x^9 + 702*x^8 + 1056*x^7 + 630*x^6 + 128*x^3 + 420*x^2 + 586*x + 318
        sage: F, dF = prod_with_derivative((f, f.derivative()) for f in fs)
        sage: F
        24*x^10 + 78*x^9 + 132*x^8 + 90*x^7 + 32*x^4 + 140*x^3 + 293*x^2 + 318*x + 135
        sage: dF
        240*x^9 + 702*x^8 + 1056*x^7 + 630*x^6 + 128*x^3 + 420*x^2 + 586*x + 318

    The main reason for this function to exist is that it allows us to
    *evaluate* the derivative of a product of polynomials at a point
    `\alpha` without ever fully expanding the product *as a polynomial*::

        sage: alpha = 42
        sage: F(alpha)
        442943981574522759
        sage: dF(alpha)
        104645261461514994
        sage: us = [f(alpha) for f in fs]
        sage: vs = [f.derivative()(alpha) for f in fs]
        sage: prod_with_derivative(zip(us, vs))
        (442943981574522759, 104645261461514994)
    """
    class _aux:
        def __init__(self, f, df):
            self.f, self.df = f, df

        def __mul__(self, other):
            return _aux(self.f * other.f, self.df * other.f + self.f * other.df)

        def __iter__(self):
            yield self.f
            yield self.df

    return tuple(prod(_aux(*tup) for tup in pairs))



# sage_setup: distribution = sagemath-categories
#*****************************************************************************
#       Copyright (C) 2010 Florent Hivert <Florent.Hivert@univ-rouen.fr>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

cpdef fibers(f, domain)

from sage.structure.parent cimport Parent
from sage.structure.list_clone cimport ClonableIntArray

cdef class FiniteSetMap_MN(ClonableIntArray):
    cpdef _setimage(self, int i, int j)
    cpdef _getimage(self, int i)
    cpdef setimage(self, i, j)
    cpdef getimage(self, i)
    cpdef domain(self)
    cpdef codomain(self)
    cpdef image_set(self)
    cpdef fibers(self)
    cpdef items(self)
    cpdef FiniteSetMap_MN _compose_internal_(self, FiniteSetMap_MN other,
                                             Parent resParent)
    cpdef check(self)

cdef class FiniteSetMap_Set(FiniteSetMap_MN): pass

cpdef FiniteSetMap_Set FiniteSetMap_Set_from_list(cls, parent, lst)
cpdef FiniteSetMap_Set FiniteSetMap_Set_from_dict(cls, parent, d)

cdef class FiniteSetEndoMap_N(FiniteSetMap_MN): pass
cdef class FiniteSetEndoMap_Set(FiniteSetMap_Set): pass
